"""Data wrangling simplicity, complete audit transparency, and at speed."""
import sys

def main():
    """Main entry point for the script."""
    pass

if __name__ == "__main__":
    sys.exit(main())


"""
Transformation functions to perform wrangling according to a defined method. Assumes the method
is valid.
"""
import pandas as pd
import numpy as np
import uuid

import whyqd.common as _c

def perform_transform(df, field_name, field_type, structure, category = None):
	"""
	A recursive transformation. A method should be a list fields upon which actions are applied, but
	each field may have nested sub-fields requiring their own actions. Before the action on the
	current field can be completed, it is necessary to perform the actions on each sub-field.

	Parameters
	----------
	df: DataFrame
		Working data to be transformed
	field_name: str
		Name of the target schema field
	structure: list
		List of fields with restructuring action defined by term 0
	category: list
		List defining unique terms to be categorised

	Returns
	-------
	Dataframe
		Containing the implementation of all nested transformations
	"""
	perform = {
		"ORDER": transform_by_order,
		"ORDER_NEW": transform_by_order_by_date,
		"ORDER_OLD": transform_by_order_by_date,
		"CALCULATE": transform_by_calculation,
		"JOIN": transform_by_joining,
		"CATEGORISE": transform_by_categorisation,
		"NEW": transform_by_new_field,
		"RENAME": transform_by_rename
	}
	# Analyse requirements for this method
	if len(structure) < 2 or structure[0]["type"] != "action":
		e = "Field `{}` has invalid method. Please review.".format(field_name)
		raise ValueError(e)
	action = structure[0]
	# Recursive check ...
	flattened_structure = [action]
	for i, field in enumerate(structure[1:]):
		if isinstance(field, list):
			# Need to create a temporary column ... the action will be performed here
			# then this nested structure will be replaced by the output of this new column
			temp_name = "nested_" + str(uuid.uuid4())
			df = perform_transform(df, temp_name, field_type, field, category=category)
			field = {
				"name": temp_name,
				"type": "nested"
			}
		flattened_structure.append(field)
	# Actions: ORDER / ORDER_NEW / ORDER_OLD / CALCULATE / JOIN / CATEGORISE / NEW / RENAME
	if action["name"] == "CATEGORISE":
		return perform[action["name"]](df, field_name, field_type,
									   flattened_structure, category=category)
	return perform[action["name"]](df, field_name, flattened_structure)

def transform_by_rename(df, field_name, structure):
	"""
	Perform a default transformation, simply renaming the field to the schema field.

	Parameters
	----------
	df: DataFrame
		Working data to be transformed
	field_name: str
		Name of the target schema field
	structure: list
		List of fields with restructuring action defined by term 0

	Returns
	-------
	Dataframe
		Containing the implementation of all nested transformations
	"""
	# Analyse requirements for this method
	if len(structure) != 2 or structure[0]["name"] != "RENAME":
		e = "Field `{}` has invalid RENAME method. Please review.".format(field_name)
		raise ValueError(e)
	# Rename, note, can only be one field if a rename ...
	# https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.rename.html
	# {'from': 'too'}
	df.rename(index=str,
			  columns= {structure[1]["name"]: field_name},
			  inplace=True)
	return df

def transform_by_new_field(df, field_name, structure):
	"""
	Add a new field to a dataframe and set its value to the default provided.

	Parameters
	----------
	df: DataFrame
		Working data to be transformed
	field_name: str
		Name of the target schema field
	structure: list
		List of fields with restructuring action defined by term 0

	Returns
	-------
	Dataframe
		Containing the implementation of all nested transformations
	"""
	if len(structure) != 2 or structure[0]["name"] != "NEW":
		e = "Field `{}` has invalid NEW method. Please review.".format(field_name)
		raise ValueError(e)
	# Add field, note, can only be one field if new ...
	df.loc[:, field_name] = structure[1]["value"]
	return df

def transform_by_joining(df, field_name, structure):
	"""
	Join a list of columns together with a space (i.e. concatenating text in multiple fields).

	Parameters
	----------
	df: DataFrame
		Working data to be transformed
	field_name: str
		Name of the target schema field
	structure: list
		List of fields with restructuring action defined by term 0

	Returns
	-------
	Dataframe
		Containing the implementation of all nested transformations
	"""
	if len(structure) < 2 or structure[0]["name"] != "JOIN":
		e = "Field `{}` has invalid JOIN method. Please review.".format(field_name)
		raise ValueError(e)
	fields = [field["name"] for field in structure[1:]]
	# https://stackoverflow.com/a/45976632
	df.loc[:, field_name] = df.loc[:, fields].apply(lambda x: "" if x.isnull().all() else
															  " ".join(x.dropna().astype(str)).strip(),
													axis=1)
	return df

def transform_by_order(df, field_name, structure):
	"""
	Create a new field by iterating over a list of fields and picking the next value in the list.

	Parameters
	----------
	df: DataFrame
		Working data to be transformed
	field_name: str
		Name of the target schema field
	structure: list
		List of fields with restructuring action defined by term 0

	Returns
	-------
	Dataframe
		Containing the implementation of all nested transformations
	"""
	if len(structure) < 2 or structure[0]["name"] != "ORDER":
		e = "Field `{}` has invalid ORDER method. Please review.".format(field_name)
		raise ValueError(e)
	fields = [field["name"] for field in structure[1:]]
	if len(fields) > 1:
		df.rename(index=str, columns= {fields[0]: field_name}, inplace=True)
		for field in fields[1:]:
			df.loc[:, field_name] = df.apply(lambda x: (x[field]
														if pd.notnull(x[field])
														else x[field_name]),
											  axis=1)
	else:
		# Deal with the single case
		structure[1]["name"] = "RENAME"
		return transform_by_rename(df, field_name, structure)
	return df

def transform_by_order_by_date(df, field_name, structure):
	"""
	Create a new field by iterating over a list of fields and picking the newest or oldest value in
	the list. Requires a "datename" constraint, otherwise it defaults to an ORDER action.

	Parameters
	----------
	df: DataFrame
		Working data to be transformed
	field_name: str
		Name of the target schema field
	structure: list
		List of fields with restructuring action defined by term 0

	Returns
	-------
	Dataframe
		Containing the implementation of all nested transformations
	"""
	if len(structure) < 4 or not structure[0]["name"].startswith("ORDER_"):
		e = "Field `{}` has invalid ORDER_BY method. Please review.".format(field_name)
		raise ValueError(e)
	action = structure[0]["name"].split("_")
	fields = [field["name"] for field in structure[1:]]
	base_date = None
	# Requires sets of 3 terms: field, +, date_field
	term_set = len(structure[0]["structure"])
	for data, modifier, date in _c.chunks(fields, term_set):
		if modifier != "+":
			e = "Field `{}` has invalid ORDER_BY method. Please review.".format(field_name)
			raise ValueError(e)
		if not base_date:
			# Start the comparison on the next round
			df.rename(index=str, columns= {data: field_name}, inplace=True)
			base_date = date
			if data == date:
				# Just comparing date columns
				base_date = field_name
			df.loc[:, base_date] = df.loc[:, base_date].apply(lambda x: pd.to_datetime(_c.parse_dates(x),
																					   errors="coerce"))
			continue
		# np.where date is <> base_date and don't replace value with null
		# logic: if test_date not null & base_date <> test_date
		# False if (test_date == nan) | (base_date == nan) | base_date >< test_date
		# Therefore we need to test again for the alternatives
		df.loc[:, date] = df.loc[:, date].apply(lambda x: pd.to_datetime(_c.parse_dates(x),
																		 errors="coerce"))
		if action[1] == "NEW":
			df.loc[:, field_name] = np.where(~pd.notnull(df[date]) | (df[base_date] > df[date]),
											  np.where(pd.notnull(df[field_name]),
													   df[field_name], df[data]),
											  np.where(pd.notnull(df[data]),
													   df[data], df[field_name]))
			if base_date != field_name:
				df.loc[:, base_date] = np.where(~pd.notnull(df[date]) | (df[base_date] > df[date]),
												np.where(pd.notnull(df[base_date]),
														 df[base_date], df[date]),
												np.where(pd.notnull(df[date]),
														 df[date], df[base_date]))
		if action[1] == "OLD":
			df.loc[:, field_name] = np.where(~pd.notnull(df[date]) | (df[base_date] < df[date]),
											  np.where(pd.notnull(df[field_name]),
													   df[field_name], df[data]),
											  np.where(pd.notnull(df[data]),
													   df[data], df[field_name]))
			if base_date != field_name:
				df.loc[:, base_date] = np.where(~pd.notnull(df[date]) | (df[base_date] < df[date]),
												np.where(pd.notnull(df[base_date]),
														 df[base_date], df[date]),
												np.where(pd.notnull(df[date]),
														 df[date], df[base_date]))
	return df

def transform_by_calculation(df, field_name, structure):
	"""
	Calculate the value of a field derived from the values of other fields. Requires a constraint
	indicating whether the fields should be ADD or SUB from the current total.

	Parameters
	----------
	df: DataFrame
		Working data to be transformed
	field_name: str
		Name of the target schema field
	structure: list
		List of fields with restructuring action defined by term 0

	Returns
	-------
	Dataframe
		Containing the implementation of all nested transformations
	"""
	if len(structure) < 3 or structure[0]["name"] != "CALCULATE":
		e = "Field `{}` has invalid CALCULATE method. Please review.".format(field_name)
		raise ValueError(e)
	# Requires sets of 2 terms: + or -, field
	term_set = len(structure[0]["structure"])
	add_fields = [field["name"] for modifier, field in _c.chunks(structure[1:], term_set)
				  if modifier["name"] == "+"]
	sub_fields = [field["name"] for modifier, field in _c.chunks(structure[1:], term_set)
				  if modifier["name"] == "-"]
	for field in add_fields + sub_fields:
		df.loc[:, field] = df.loc[:, field].apply(lambda x: _c.parse_float(x))
	df.loc[:, field_name] = df[add_fields].abs().sum(axis=1) - df[sub_fields].abs().sum(axis=1)
	return df

def transform_by_categorisation(df, field_name, field_type, structure, category = None):
	"""
	Produce categories from terms or headers. There are three categorisation options:

		1. Term-data are categories derived from values in the data,
		2. Header-data are terms derived from the header name and boolean True for any value,
		3. "Boolean"-data, the category itself is True/False. Default in a boolean is True.

	Categorisation is a special case, requiring both method fields, and method categories, along
	with the schema field_name field to lookup the required term definitions.

	Parameters
	----------
	df: DataFrame
		Working data to be transformed
	field_name: str
		Name of the target schema field
	structure: list
		List of fields with restructuring action defined by term 0
	category: list
		List defining unique terms to be categorised

	Returns
	-------
	Dataframe
		Containing the implementation of all nested transformations
	"""
	if len(structure) < 2 or structure[0]["name"] != "CATEGORISE" or not category:
		e = "Field `{}` has invalid CATEGORISE method. Please review.".format(field_name)
		raise ValueError(e)
	# Fields are in sets of 2 terms: + or -, field
	# The modifier defines one of two approaches:
	#   '+': The terms in the field are used to identify the schema category
	#   '-': Non-null terms indicate presence of a schema category
	# If field type is 'array', then the destination terms are lists;
	# If field type is 'boolean', then the default term is True;
	is_array = True if field_type == "array" else False
	is_boolean = True if field_type == "boolean" else False
	# Sort out boolean term names before they cause further pain, correct later ...
	if is_boolean:
		for c in category:
			if c["name"]: c["name"] = "true"
			else: c["name"] = "false"
	default = "None" if is_array else is_boolean
	# Set the field according to the default
	#https://stackoverflow.com/a/31469249
	df[field_name] = [[] for _ in range(len(df))] if is_array else default
	# Develop the terms and conditions to assess membership of a category
	# Requires sets of 2 terms: + or -, field
	new_field = []
	term_set = len(structure[0]["structure"])
	all_terms = [c["name"] for c in category]
	# Any fields modified below must be restored: (tmp_column, original_column)
	modified_fields = []
	for modifier, field in _c.chunks(structure[1:], term_set):
		# https://docs.scipy.org/doc/numpy/reference/generated/numpy.select.html
		# Extract only the terms valid for this particular field
		terms = []
		for field_term in all_terms:
			for f in category:
				if f["name"] == field_term:
					for c in f["category_input"]:
						if c["column"] == field["name"]:
							terms.extend([field_term for t in c["terms"]])
					break
		if modifier["name"] == "+":
			conditions = [df[field["name"]].isin([t for subterms in
												  [item["terms"] for category_input in category
												   for item in category_input["category_input"]
												   if category_input["name"] == field_term]
												  for t in subterms])
						  for field_term in terms]
		else:
			# Modifier is -, so can make certain assumptions
			# - Terms are categorised as True or False, so choices are only True or False
			# - However, all terms allocated to 'true' will fail (since True on default is True no matter)
			# - User may return both True / False or only one
			# - If both True and False, keep False. If True, convert the True to False
			# Create temporary column to preserve data
			if df[field["name"]].dtype in ["float64", "int64", "datetime64[ns]"]:
				tmp = "tmp_" + str(uuid.uuid4())
				modified_fields.append((tmp, field["name"]))
				df[tmp] = df[field["name"]].copy()
			# Ensure any numerical zeros are nan'ed +
			if df[field["name"]].dtype in ["float64", "int64"]:
				df[field["name"]] = df[field["name"]].replace({0:np.nan, 0.0:np.nan})
			if df[field["name"]].dtype in ["datetime64[ns]"]:
				df.loc[:, field["name"]] = df.loc[:, field["name"]].apply(lambda x: pd.to_datetime(_c.parse_dates(x),
																								   errors="coerce"))
			conditions = [pd.notnull(df[field["name"]])
						  if category["fields"][field_term]["fields"][0]["name"] else
						  ~pd.notnull(df[field["name"]])
						  for field_term in terms]
		if is_boolean and modifier["name"] == "-":
			# if len(terms) == 1 and 'false', do nothing; if 'true', invert;
			# if len(terms) == 2, use 'false' only
			if "true" in terms:
				# We need to correct for the disaster of 'true'
				invrt = dict(zip(terms, conditions))
				if len(terms) == 1:
					invrt["false"] = ~invrt["true"]
				terms = [t for t in invrt.keys() if t != "true"]
				conditions = [invrt["false"]]
		# Only two terms, True or False. Reset the dictionary names
		if is_boolean and "false" in terms:
			terms = [False if t == "false" else t for t in terms]
		if not is_array:
			# Set the field terms immediately for membership, note, if no data defaults to current
			# i.e. this is equivalent to order, but with category data
			df[field_name] = np.select(conditions, terms, default=df[field_name])
		else:
			if terms and conditions:
				new_field.append(np.select(conditions, terms, default="none").tolist())
	# If a list of terms, organise the nested lists and then set the new field
	if is_array:
		# Sorted to avoid hashing errors later ...
		new_field = [sorted(list(set(x))) for x in zip(*new_field)]
		for n in new_field:
			if "none" in n: n.remove("none")
		if new_field:
			df[field_name] = new_field
	# Finally, fix the artifact columns introduced in the dataframe in case these are used elsewhere
	for tmp, original in modified_fields:
		df[original] = df[tmp].copy()
		del df[tmp]
	return df


"""
.. module:: schema
   :synopsis: Create and manage a target data schema.

.. moduleauthor:: Gavin Chait <github.com/turukawa>

Schema
======

Creating a `schema` is the first part of the wrangling process. Your schema defines the structural
metadata target for your wrangling process. This is not the format your input data arrive in, but
it is what you require it to look like when you're done.

Your schema sets the requirements, constraints and sensible defaults available for creating each
`method` that will describe the process for wrangling input data into the fields defined by the
schema. It can be reused to create multiple `methods`. Once complete, automated scripts can perform
further cleaning and validation.

In simple terms, the columns in an input CSV or Excel-file will be restructured into new columns
defined by the fields in your schema. These target fields are likely to be those in your database,
or in your analytical software. Until your input data conform to this structure, you can't do your
real work.

Minimum valid requirements
--------------------------

A minimum valid schema requires a `name` to identify the schema, and a single, minimally-valid
`field` containing a `name` and `type`::

	{
	  "name": "A simple name",
	  "fields": [
		{
			"name": "Field name, e.g. 'column_name'",
			"type": "Valid data type, e.g. 'string', 'number'"
		}
	  ]
	}

Everything else is optional, unless specifically required by that field-type.

Schema descriptors
------------------

Schema terms include:

`name`
^^^^^^
This is a required term. Spaces will be replaced with `_` and the string will be lowercased.

`title`
^^^^^^^
A human-readable version of the schema name.

`description`
^^^^^^^^^^^^^
A complete description of the schema. Depending on how complex your work becomes, try and be as
helpful as possible to 'future-you'. You'll thank yourself later.

Field descriptors
-----------------

Fields, similarly, contain `name`, `title` and `description`, as well as `type` as compulsory. To
see the available field types::

	>>> sc = _d.Schema()
	>>> sc.default_field_types
	['string',
	'number',
	'integer',
	'boolean',
	'object',
	'array',
	'date',
	'datetime',
	'year']

To see further parameter options for each default type::

	>>> sc.default_field_settings('string')
	{
		'required': ['name', 'type'],
		'name': 'field_name',
		'title': 'A human-readable version of the field name',
		'description': 'Any text-based string.',
		'type': 'string',
		'format': {
			'description': 'The format keyword options for `string` are `default`, `email`, `uri`, `binary`, and `uuid`.',
			'category': ['default', 'email', 'uri', 'binary', 'uuid'],
			'default': 'default'
			},
		'foreignKey': {
			'type': 'boolean',
			'description': 'Set `foreignKey` `true` if the field is to be treated as an immutable value.'
			},
		'constraints': {
			'description': 'The following constraints are supported.',
			'properties': {
				'required': {
					'type': 'boolean',
					'description': 'Indicates whether a property must have a value for each instance.'
				},
				'unique': {
					'type': 'boolean',
					'description': 'When `true`, each value for the property `MUST` be unique.'
				},
			'category': {
				'type': 'array',
				'minItems': 1,
				'uniqueItems': True,
				'terms': {
					'type': 'string'
					}
				},
			'minimum': {
				'type': 'integer',
				'description': 'An integer that specifies the minimum length of a value.'
				},
			'maximum': {
				'type': 'integer',
				'description': 'An integer that specifies the maximum length of a value.'
				}
			}
		},
	'missing': {
		'description': 'Default to be used for missing values.',
		'default': ''
		}
	}

`name`
^^^^^^
This is a required term and is equivalent to a column header. Spaces will be replaced with `_` and
the string will be lowercased.

`title`
^^^^^^^
A human-readable version of the field name.

`description`
^^^^^^^^^^^^^
A complete description of the field. As for the schema, try and be as helpful as possible to
future-you.

`foreignKey`
^^^^^^^^^^^^
This is a boolean term, only required if you need this field to be treated as a foreign-key or
identifier for your destination data::

	"foreignKey": True

Data in this field will not be tested for uniqueness. Instead, these data will remain immutable, not
being 'forced' into a date or number type to preserve whatever fruity formatting are described in
your input data.

During the wrangling process, this field can be used for merging with other input
data, ensuring consistency between sources.

`type` and `format`
^^^^^^^^^^^^^^^^^^^
`type` defines the data-type of the field, while `format` - which is currently unsupported in
wrangling - further refines the specific `type` properties. The core supported types, with indents
for formats:

* `string`: any text-based string.

  * `default`: any string
  * `email`: an email address
  * `uri`: any web address / URI

* `number`: any number-based value, including integers and floats.
* `integer`: any integer-based value.
* `boolean`: a boolean [`true`, `false`] value. Can set category constraints to fix term used.
* `object`: any valid JSON data.
* `array`: any valid array-based data.
* `date`: any date without a time. Must be in ISO8601 format, `YYYY-MM-DD`.
* `datetime`: any date with a time. Must be in ISO8601 format, with UTC time specified (optionally) as `YYYY-MM-DD hh:mm:ss Zz`.
* `year`: any year, formatted as `YYYY`.

`missing`
^^^^^^^^^
`missing` defines the value to be used for any blank values in a column. This is normally `""` for
text and `np.nan` for numbers or dates. You can, however, set your own defaults for each field.

Field constraints
-----------------
`Constraints` are optional parameters that refine input data wrangling and act as a primary form of
validation. Not all of these are available to every `type`, and `default_field_settings(type)`
will list constraints available to a specific field type.

Define these as part of your schema definition for a specific field::

	{
	  "name": "A simple name",
	  "fields": [
		{
			"name": "Field name, e.g. 'column_name'",
			"type": "Valid data type, e.g. 'string', 'number'",
			"constraints": {
				"required": True,
				"unique": True
			}
		}
	  ]
	}

All available constraints:

* `required`: boolean, indicates whether this field is compulsory (but blank values in the input column are permitted and will be set to the `missing` default)
* `unique`: boolean, if `true` then all values for that input column must be unique
* `minimum`: `integer` / `number`, as appropriate defining min number of characters in a string, or the min values of numbers or integers
* `maximum`: `integer` / `number`, as appropriate defining max number of characters in a string, or the max values of numbers or integers

Field constraints: category
---------------------------
In FrictionlessData.io, this is called `enum`, which isn't particularly meaningful. `Category` data
are the set of unique category terms permitted in this field. During wrangling you will be able to
define values which should be assigned to each of these categories.

Define these as part of your schema definition for a specific field::

	{
	  "name": "A simple name",
	  "fields": [
		{
			"name": "Field name, e.g. 'column_name'",
			"type": "Valid data type, e.g. 'string', 'number'",
			"constraints": {
				"category": ["cheddar", "gouda", "other"]
			}
		}
	  ]
	}

Each field `type` will have its own category constraints. For example, boolean categories can use a
different term than True / False defined by the category, but only permits two terms. Others have
a minimum of one term in a category, but require the list member type to be `string`, `number`, etc.
Ordinarily, `category` terms must be unique.

Review the `default_field_settings(type)` for that field's specific category constraints.

Field constraints: filter
--------------------------
`Filters` are a constraint that filter a named field, or the `foreignKey`, by date-limited data.

Define these as part of your schema definition for a valid field::

	{
	  "name": "A simple name",
	  "fields": [
		{
			"name": "Field name, e.g. 'column_name'",
			"type": "Valid data type, e.g. 'date', 'datetime'",
			"filter": {
				"field": "foreignKey",
				"modifiers": ["LATEST", "AFTER"]
			}
		}
	  ]
	}

There are two compulsory parameters defining a filter:

* `field`: another field which is the subject of this filter, or by default the 'foreignKey'.
* `modifiers`: an array of permitted filter terms, including any of ["LATEST", "AFTER", "BEFORE", "ALL"].

call `default_filter_names()` to get a list, and `default_filter_settings(filter_name)` to get a
definition

For example, to filter all foreign keys (which may be duplicated as part of a time-series) to be
more recent than a specified date, include "AFTER" in your list of filter modifiers.
"""

from copy import deepcopy

from whyqd.schema import Field
import whyqd.common as _c

class Schema:
	"""Create and manage a target schema for a wrangling process.

	Parameters
	----------
	source: path to a json file containing a saved schema, default is None
	kwargs: a schema defined as a dictionary, or default blank dictionary
	"""
	def __init__(self, source=None, **kwargs):
		self.required_field_terms = ["name", "type"]
		# Get default fields
		self.default_fields = self.build_default_fields()
		self.default_filters = _c.get_settings("filter")
		if source:
			kwargs = _c.load_json(source)
		self.schema_settings = deepcopy(kwargs)
		self.build()

	def __repr__(self):
		"""
		Returns the string representation of the model.
		"""
		if self.schema_settings.get("name"):
			return "Schema: `{}`".format(self.schema_settings["name"])
		return "Schema"

	@property
	def details(self):
		"""
		Schema name, title and description.

		Parameters
		----------
		name: string
			Term used for filename and referencing. Will be lower-cased and spaces replaced with `_`
		title: string
			Human-readable term used as name.
		description: string
			Detailed description for the schema. Reference its objective and use-case.

		Returns
		-------
		dict
		"""
		response = {
			"name": self.schema_settings.get("name"),
			"title": self.schema_settings.get("title"),
			"description": self.schema_settings.get("description")
		}
		return response

	def set_details(self, name=None, title=None, description=None):
		"""
		Set schema name, title and description. Can also be used to update existing information.

		Parameters
		----------
		name: string
			Term used for filename and referencing. Will be lower-cased and spaces replaced with `_`
		title: string
			Human-readable term used as name.
		description: string
			Detailed description for the schema. Reference its objective and use-case.

		Raises
		------
		KeyError: if missing name
		"""
		if "name" not in self.schema_settings and not name:
			# `name` required
			e = "`name` not found in target schema"
			raise KeyError(e)
		for v in [name, title, description]:
			if v and not isinstance(v, str):
				e = "`{}` is not a valid string. Schema details must be strings."
				raise TypeError(e)
		self.schema_settings["name"] = "_".join(name.split(" ")).lower()
		if title: self.schema_settings["title"] = title
		if description: self.schema_settings["description"] = description

	@property
	def default_filter_names(self):
		"""
		Default list of filter names available as field constraints for the schema. Returns only a
		list of types. Details for individual filters can be returned with
		`default_filter_settings`.

		Returns
		-------
		list
		"""
		return [f["name"] for f in self.default_filters["filter"]["modifiers"]]

	def default_filter_settings(self, filter_name):
		"""
		Get the default settings available for a specific filter.

		Parameters
		----------
		filter_name: string
			A specific name for a filter type (as listed in `default_filter_names`).

		Returns
		-------
		dict, or empty dict if no such `filter_name`
		"""
		for filter in self.default_filters["filter"]["modifiers"]:
			if filter["name"] == filter_name:
				return deepcopy(filter)
		return {}

	@property
	def default_field_types(self):
		"""
		Default list of field names available to define fields for the schema. Returns only a list
		of types. Details for individual default fields can be returned with
		`default_field_settings`.

		Returns
		-------
		list
		"""
		return [f.type for f in self.default_fields]

	def default_field_settings(self, field_type):
		"""
		Get the default settings available for a specific field type.

		Parameters
		----------
		field_type: string
			A specific term for a field type (as listed in `default_field_types`).

		Returns
		-------
		dict, or empty dict if no such `field_type`
		"""
		for f in self.default_fields:
			if f.type == field_type:
				return deepcopy(f.settings)
		return {}

	@property
	def all_fields(self):
		"""
		Get dictionary of fields, where key: `field_name` and value: `type`.

		Returns
		-------
		dict, or empty dict if no fields
		"""
		return {f.name: f.type for f in self.schema_settings.get("fields", [])}

	@property
	def all_field_names(self):
		"""
		Get list of field names,.

		Returns
		-------
		list
		"""
		return [f.name for f in self.schema_settings.get("fields", [])]

	def field(self, name):
		"""
		A specific field from the list of fields defining this schema, called by a unique `name`.

		Parameters
		----------
		name: string
			Field names must be unique, so a valid `name` in the field list will have no collisions.

		Returns
		-------
		dict, or empty dict if no such `name`
		"""
		for field in self.schema_settings.get("fields", []):
			if field.name == name:
				return deepcopy(field.settings)
		e = "`{}` not found in Schema fields.".format(name)
		raise ValueError(e)

	def set_field(self, **kwargs):
		"""
		Set the parameters for a specific field to define this schema, called by a unique `name`.
		If the `name` is already in the schema, then this will update that field.

		Parameters
		----------
		kwargs: dict
			Parameters will be validated against the type requirements, so check carefully with
			`default_field_settings`.
		"""
		kwargs = deepcopy(kwargs)
		if "name" not in kwargs:
			e = "No valid `name` provided for Field."
			raise ValueError(e)
		self.schema_settings.get("fields", [])[:] = [f for f in self.schema_settings.get("fields", [])
													 if f.name != kwargs["name"]]
		kwargs["name"] = "_".join(kwargs["name"].split(" ")).lower()
		field = self.build_field(**kwargs)
		self.schema_settings["fields"].append(field)

	def set_field_constraints(self, name, **constraints):
		"""
		Set the constraint parameters for a specific field to define this schema, called by a unique
		`name` already in the schema.

		The structure of the constraints is defined as follows::

			{
				"key": "value",
				"key": "value"
			}

		`category` is a special constraint that can be defined as e.g.::

			{
				"category": ["term1", "term2"]
			}

		All that is required is a list, and the function will take care of the formal structure.
		`filter` are managed in the Method part of the process (since this are defined in respect
		to the data being structured).

		Parameters
		----------
		name: string
			Specific name for a field already in the Schema
		constraints: dict
			A set of key:value pairs defining constraints as described in `default_field_settings`.
		"""
		field = self.field(name)
		valid_constraints = self.default_field_settings(field["type"]).get("constraints", {}).keys()
		if not valid_constraints:
			e = "Field `{}` has no permitted constraints.".format(name)
			raise ValueError(e)
		for constraint in constraints:
			if constraint not in valid_constraints:
				e = "Constraint `{}` is not permitted in field `{}`.".format(constraint, name)
				raise KeyError(e)
		category = constraints.pop("category", None)
		if not field.get("constraints"):
			field["constraints"] = {}
		if constraints:
			field["constraints"] = {**field["constraints"], **constraints}
			self.set_field(**field)
		if category: self.set_field_category(name, *category["category"])

	def set_field_category(self, name, *categories, overwrite=True):
		"""
		Set the category constraint parameters for a specific field. `categories` is defined as e.g.::

			["term1", "term2"]

		All that is required is a list, and the function will take care of the formal structure.

		Parameters
		----------
		name: string
			Specific name for a field already in the Schema
		categories: list
			A list of string terms defining target categories.
		overwrite: bool
			If field has existing category constraints, then overwrite with this new list.
		"""
		field = self.field(name)
		if not overwrite:
			original = [c["name"] for c in field.get("constraints", {}).get("category", [])]
			categories.extend(original)
			categories = set(categories)
		category = [{"name": c} for c in categories]
		if not field.get("constraints"):
			field["constraints"] = {}
		field["constraints"]["category"] = category
		self.set_field(**field)

	def set_field_filters(self, name, *filters):
		"""
		Set the filter parameters for a specific field to define this schema, called by a unique
		`name` already in the schema.

		Parameters
		----------
		name: string
			Specific name for a field already in the Schema
		filters: list of strings
			A list of filter names as described in `default_filter_names`.
		"""
		field = self.field(name)
		self.schema_settings["fields"][:] = [f for f in self.schema_settings["fields"]
											 if f.name != field.name]
		# update field
		filters = deepcopy(filters)
		modifiers = [f for f in self.default_filters["filter"]["modifiers"]
					 if f["name"] in filters]
		if modifiers:
			filters = {
				"field": self.default_filters["filter"]["field"],
				"modifiers": modifiers
			}
			field["constraints"]["properties"]["filter"] = filters
			field = self.build_field(**field)
			self.schema_settings["fields"].append(field)

	def build_field(self, **field):
		"""
		For a list of fields, defined as dictionaries, create and return Field objects.

		Parameters
		----------
		field: dictionary of Field parameters

		Raises
		------
		ValueError: if field fails validation

		Returns
		-------
		Field
		"""
		field = deepcopy(field)
		if not isinstance(field, dict) or not all(key in field for key in self.required_field_terms):
			e = "Field is not a valid dictionary"
			raise ValueError(e)
		_name = field["name"]
		del field["name"]
		_type = field["type"]
		del field["type"]
		field = Field(_name, _type, **field)
		if not field.validates:
			e = "Field `{}` of type `{}` does not validate".format(_name, _type)
			raise ValueError(e)
		return field

	def build(self):
		"""
		Build and validate the Schema.
		"""
		self.schema_settings["fields"] = [self.build_field(**field) for field in
										  self.schema_settings.get("fields", [])]

	def build_default_fields(self):
		"""
		Build the default fields for presentation to the user as options.

		Returns
		-------
		list of Fields
		"""
		default_fields = _c.get_settings("schema")
		default_fields = [{**default_fields["common"], **f}
						  for f in default_fields["fields"]]
		response = []
		for field in default_fields:
			_name = field["name"]
			del field["name"]
			_type = field["type"]
			del field["type"]
			response.append(Field(_name, _type, validate=False, **field))
		return response

	@property
	def validates(self):
		"""
		Schema validates with all fields unique and required terms.

		Raises
		------
		ValueError on field failure.

		Returns
		-------
		bool: True for validates
		"""
		if not all([field.validates for field in self.schema_settings.get("fields", [])]):
			e = "Individual `fields` in schema are not valid."
			raise ValueError(e)
		fields = [field.name for field in self.schema_settings.get("fields", [])]
		if not fields:
			e = "No `fields` in schema. Need at least one."
			raise ValueError(e)
		if len(fields) != len(set(fields)):
			e = "Schema `field` names not unique: `{}`.".format(fields)
			raise ValueError(e)
		if not "name" in self.schema_settings:
			e = "Schema `name` not present."
			raise ValueError(e)
		return True

	@property
	def settings(self):
		"""
		Schema settings returned as a dictionary.

		Returns
		-------
		dict: settings
		"""
		schema_dict = deepcopy(self.schema_settings)
		schema_dict["fields"] = [field.settings for field in self.schema_settings.get("fields", [])]
		return schema_dict

	def save(self, directory, filename=None, overwrite=False, created_by=None):
		"""
		Schema settings returned as a dictionary.

		Parameters
		----------
		directory: the destination directory
		filename: default to schema name
		overwrite: bool, True if overwrite existing file
		created_by: string, or None, to define the schema creator/updater

		Raises
		------
		ValueError if no `filename`

		Returns
		-------
		bool True if saved
		"""
		if not filename: filename = self.schema_settings.get("name")
		if not filename:
			e = "Schema save requires a valid filename"
			raise ValueError(e)
		if filename.split(".")[-1] != "json":
			filename += ".json"
		data = self.settings
		now = _c.get_now()
		if "created" not in data:
			data["created"] = now
		data["updated"] = now
		if created_by and isinstance(created_by, str):
			if "created_by" not in data:
				data["created_by"] = created_by
			data["updated_by"] = created_by
		return _c.save_json(data, directory + filename, overwrite)


"""
.. module:: method
   :synopsis: Create and manage a wrangling method based on a predefined schema.

.. moduleauthor:: Gavin Chait <github.com/turukawa>

Method
======

Once you have created your `Schema` it can be imported and used to develop a wrangling `method`, a
complete, structured JSON file which describes all aspects of the wrangling process. There is no
'magic'. Only what is defined in the method will be executed during transformation.

A method file can be shared, along with your input data, and anyone can import **whyqd** and
validate your method to verify that your output data is the product of these inputs::

	import whyqd as _w
	method = _w.Method(source, directory=DIRECTORY, input_data=INPUT_DATA)

`source` is the full path to the schema you wish to use, and `DIRECTORY` will be your working
directory for saved data, imports, working data and output.

`INPUT_DATA` is a list of filenames or file sources. This is optional at this stage, and you can
add and edit your sources later. File sources can include URI's.

Help
----

To get help, type::

	method.help()
	# or
	method.help(option)

Where `option` can be any of::

	"status"
	"merge"
	"structure"
	"category"
	"filter"

`status` will return the current method status, and your mostly likely next steps. The other options
will return methodology, and output of that option's result (if appropriate).

These are the steps to create a complete method:

Merge
-----
`merge` will join, in order from right to left, your input data on a common column. You can modify
your input data at any time. Note, however, that this will reset your status and require
revalidation of all subsequent steps.

To add input data, where `input_data` is a filename / source, or list of filenames / sources::

	method.add_input_data(input_data)

To remove input data, where `id` is the unique id for that input data:

	method.remove_input_data(id)

To display a nicely-formatted output for review::

	# Permits horizontal scroll-bar in Jupyter Notebook
	from IPython.core.display import HTML
	display(HTML("<style>pre { white-space: pre !important; }</style>"))

	print(method.pretty_print_input_data())

	Data id: c8944fed-4e8c-4cbd-807d-53fcc96b7018

	====  ====================  ========================  =================================  =============================  =======================================================  =============================  ===========================
	  ..  Account Start date      Current Rateable Value  Current Relief Award Start Date    Current Relief Type            Full Property Address                                    Primary Liable party name        Property Reference Number
	====  ====================  ========================  =================================  =============================  =======================================================  =============================  ===========================
	   0  2003-05-14 00:00:00                       8600  2019-04-01 00:00:00                Retail Discount                Ground Floor, 25, Albert Road, Southsea, Hants, PO5 2SE  Personal details not supplied                 177500080710
	   1  2003-07-28 00:00:00                       9900  2005-04-01 00:00:00                Small Business Relief England  Ground Floor, 102, London Road, Portsmouth, PO2 0LZ      Personal details not supplied                 177504942310
	   2  2003-07-08 00:00:00                       6400  2005-04-01 00:00:00                Small Business Relief England  33, Festing Road, Southsea, Hants, PO4 0NG               Personal details not supplied                 177502823510
	====  ====================  ========================  =================================  =============================  =======================================================  =============================  ===========================

	Data id: a9ad7716-f777-4752-8627-dd6206bede65

	====  ===================================  =================================  ========================  ================================================================  =======================================================  ===========================
	  ..  Current Prop Exemption Start Date    Current Property Exemption Code      Current Rateable Value  Full Property Address                                             Primary Liable party name                                  Property Reference Number
	====  ===================================  =================================  ========================  ================================================================  =======================================================  ===========================
	   0  2019-11-08 00:00:00                  LOW RV                                                  700  Advertising Right, 29 Albert Road, Portsmouth, PO5 2SE            Personal details not supplied                                           177512281010
	   1  2019-09-23 00:00:00                  INDUSTRIAL                                            11000  24, Ordnance Court, Ackworth Road, Portsmouth, PO3 5RZ            Personal details not supplied                                           177590107810
	   2  2019-09-13 00:00:00                  EPRI                                                  26500  Unit 12, Admiral Park, Airport Service Road, Portsmouth, PO3 5RQ  Legal & General Property Partners (Industrial Fund) Ltd                 177500058410
	====  ===================================  =================================  ========================  ================================================================  =======================================================  ===========================

	Data id: 1e5a165d-5e83-4eec-9781-d450a1d3f5f1

	====  ====================  ========================  =========================================================================  ==========================================  =================
	  ..  Account Start date      Current Rateable Value  Full Property Address                                                      Primary Liable party name                     Property ref no
	====  ====================  ========================  =========================================================================  ==========================================  =================
	   0  2003-11-10 00:00:00                      37000  Unit 7b, The Pompey Centre, Dickinson Road, Southsea, Hants, PO4 8SH       City Electrical Factors  Ltd                     177200066910
	   1  2003-11-08 00:00:00                     594000  Express By Holiday Inn, The Plaza, Gunwharf Quays, Portsmouth, PO1 3FD     Kew Green Hotels (Portsmouth Lrg1) Limited       177209823010
	   2  1994-12-25 00:00:00                      13250  Unit 2cd, Shawcross Industrial Estate, Ackworth Road, Portsmouth, PO3 5JP  Personal details not supplied                    177500013310
	====  ====================  ========================  =========================================================================  ==========================================  =================

Once you're satisfied with your `input_data`, prepare an `order_and_key` list to define the merge
order, and a unique key for merging. Each input data file needs to be defined in a list as a dict::

	{id: input_data id, key: column_name for merge}

Run the merge by calling (and, optionally - if you need to overwrite an existing merge - setting
`overwrite_working=True`)::

	method.merge(order_and_key, overwrite_working=True)

To view your existing `input_data` as a JSON output (or the `pretty_print_input_data` as above)::

	method.input_data

Structure
---------
`structure` is the core of the wrangling process and is the step where you define the actions
which must be performed to restructure your working data.

Create a list of methods of the form::

	{
		"schema_field1": ["action", "column_name1", ["action", "column_name2"]],
		"schema_field2": ["action", "column_name1", "modifier", ["action", "column_name2"]],
	}

The format for defining a `structure` is as follows, and - yes - this does permit you to create
nested wrangling tasks::

	[action, column_name, [action, column_name]]

e.g.::

	["CATEGORISE", "+", ["ORDER", "column_1", "column_2"]]

This permits the creation of quite expressive wrangling structures from simple building blocks.

Every task structure must start with an action to describe what to do with the following terms.
There are several "actions" which can be performed, and some require action modifiers:

	* NEW: Add in a new column, and populate it according to the value in the "new" constraint

	* RENAME: If only 1 item in list of source fields, then rename that field

	* ORDER: If > 1 item in list of source fields, pick the value from the column, replacing each value with one from the next in the order of the provided fields

	* ORDER_NEW: As in ORDER, but replacing each value with one associated with a newer "dateorder" constraint

		* MODIFIER: `+` between terms for source and source_date

	* ORDER_OLD: As in ORDER, but replacing each value with one associated with an older "dateorder" constraint

		* MODIFIER: `+` between terms for source and source_date

	* CALCULATE: Only if of "type" = "float64" (or which can be forced to float64)

		* MODIFIER: `+` or `-` before each term to define whether add or subtract

	* JOIN: Only if of "type" = "object", join text with " ".join()

	* CATEGORISE: Only if of "type" = "string"; look for associated constraint, "categorise" where `True` = keep a list of categories, `False` = set True if terms found in list

		* MODIFIER:

			* `+` before terms where column values to be classified as unique

			* `-` before terms where column values are treated as boolean

Category
--------

Provide a list of categories of the form::

	{
		"schema_field1": {
			"category_1": ["term1", "term2", "term3"],
			"category_2": ["term4", "term5", "term6"]
		}
	}

The format for defining a `category` term as follows::

	term_name::column_name

Get a list of available terms, and the categories for assignment, by calling::

	method.category(field_name)

Once your data are prepared as above::

	method.set_category(**category)

Filter
------
Set date filters on any date-type fields. **whyqd** offers only rudimentary post-
wrangling functionality. Filters are there to, for example, facilitate importing data
outside the bounds of a previous import.

This is also an optional step. By default, if no filters are present, the transformed output
will include `ALL` data. Parameters for filtering:

	* `field_name`: Name of field on which filters to be set
	* `filter_name`: Name of filter type from the list of valid filter names
	* `filter_date`: A date in the format specified by the field type
	* `foreign_field`: Name of field to which filter will be applied. Defaults to `field_name`

There are four filter_names:

	* `ALL`: default, import all data
	* `LATEST`: only the latest date
	* `BEFORE`: before a specified date
	* `AFTER`: after a specified date

`BEFORE` and `AFTER` take an optional `foreign_field` term for filtering on that column. e.g::

	method.set_filter("occupation_state_date", "AFTER", "2019-09-01", "ba_ref")

Filters references in column `ba_ref` by dates in column `occupation_state_date` after `2019-09-01`.

Validation
----------
Each step can be validated and, once all steps validate, you can move to transformation of your data::

	method.validate_input_data
	method.validate_merge_data
	method.validate_merge
	method.validate_structure
	method.validate_category
	method.validate_filter

Or, to run all the above and complete the method (setting status to 'Ready to Transform')::

	method.validate

Transform
---------

Transformation requires only the following::

	method.transform()
	method.save(DIRECTORY, filename=FILENAME, overwrite=True)

With one little permutation ... if you've ever created a transform before, you'll need to deliberately
tell the function to overwrite your original::

	method.transform(overwrite_output=True)

Citation
--------

**whyqd** is designed for sharing. Add information you wish to be cited to a `constructor` field in the `method`.

The `constructor` field is there to store any metadata you wish to add. Whether it be `Dublin Core <https://dublincore.org/>`_
or `SDMX <https://sdmx.org/>`_, add that metadata by creating a dictionary and placing it in the
`constructor`.

A citation is a special set of fields, with options for:

* **authors**: a list of author names in the format, and order, you wish to reference them
* **date**: publication date
* **title**: a text field for the full study title
* **repository**: the organisation, or distributor, responsible for hosting your data (and your method file)
* **doi**: the persistent `DOI <http://www.doi.org/>`_ for your repository

Those of you familiar with Dataverse's `universal numerical fingerprint <http://guides.dataverse.org/en/latest/developers/unf/index.html>`_
may be wondering where it is? **whyqd**, similarly, produces a unique hash for each datasource,
including inputs, working data, and outputs. Ours is based on `BLAKE2b <https://en.wikipedia.org/wiki/BLAKE_(hash_function)>`_
and is sufficiently universally available as to ensure you can run this as required.

As an example::

	citation = {
		"authors": ["Gavin Chait"],
		"date": "2020-02-18",
		"title": "Portsmouth City Council normalised database of commercial ratepayers",
		"repository": "Github.com"
	}
	method.set_constructors({"citation": citation})
	method.save(DIRECTORY, filename=FILENAME, overwrite=True)

You can then get your citation report::

	method.citation

	Gavin Chait, 2020-02-18, Portsmouth City Council normalised database of commercial ratepayers,
	Github.com, 1367d4f02c99030f6645389141b85a93d54c226b435fb1b5a6cbccd7f703687e442a011f62c1381793a2d3fbf13cc52c176e0c5c573008991134658759eef948,
	[input sources:
	https://www.portsmouth.gov.uk/ext/documents-external/biz-ndr-properties-january-2020.xls,
	476089d8f37581613344873068d6e94f8cd63a1a64b421edf374a2b341bc7563aff03b86db4d3fec8ca90ce150ba1e531e3ff0d374f932d13fc103fd709e01bd;
	https://www.portsmouth.gov.uk/ext/documents-external/biz-ndr-reliefs-january-2020.xls,
	892ec5b6e9b1f68e0b371bbaed8d93095d57f2b656753af2b279aee17b5854c5e9d731b2795aac285d7f7d9f5991311bc8fae0cfe5446a47163f30f0314cac06;
	https://www.portsmouth.gov.uk/ext/documents-external/biz-empty-commercial-properties-january-2020.xls,
	a41b4eb629c249fd59e6816d10d113bf2b9594c7dd7f9a61a82333a8a41bf07e59f9104eb3c1dc4269607de5a4a12eaf3215d0afc7545fdb1dfe7fe1bf5e0d29]
"""
import sys, os, uuid
from shutil import copyfile, SameFileError
import urllib.request
from copy import deepcopy
import pandas as pd
import numpy as np
from tabulate import tabulate
from operator import itemgetter

try:
	assert sys.version_info >= (3,7)
except AssertionError:
	from whyqd import common as _c
	from whyqd.method import transform as task
else:
	import whyqd.common as _c
	import whyqd.method.transform as task
from whyqd.schema import Schema
from whyqd.method import Action


STATUS_CODES = {
	"WAITING": "Waiting ...",
	"PROCESSING": "Processing ...",
	"READY_MERGE": "Ready to Merge",
	"READY_STRUCTURE": "Ready to Structure",
	"READY_CATEGORY": "Ready to Categorise",
	"READY_FILTER": "Ready to Filter",
	"READY_TRANSFORM": "Ready to Transform",
	"CREATE_ERROR": "Create Error",
	"MERGE_ERROR": "Merge Error",
	"STRUCTURE_ERROR": "Structure Error",
	"CATEGORY_ERROR": "Categorisation Error",
	"TRANSFORMATION_ERROR": "Transform Error",
	"PROCESS_COMPLETE": "Process Complete"
}

class Method(Schema):
	"""
	Create and manage a method to perform a wrangling process.

	Parameters
	----------
	source: path to a json file containing a saved schema, default is None
	directory: working path for creating methods, interim data files and final output
	kwargs: a schema defined as a dictionary, or default blank dictionary
	"""
	def __init__(self, source=None, **kwargs):
		self._status = "WAITING"
		# Clear kwargs of things we need to process prior to initialising
		self.directory = kwargs.pop("directory", _c.get_path())
		if not self.directory.endswith("/"):
			self.directory += "/"
		_c.check_path(self.directory)
		constructors = kwargs.pop("constructors", None)
		input_data = kwargs.pop("input_data", None)
		super().__init__(source=source, **kwargs)
		# Initialise after Schema initialisation
		self.default_actions = self.build_default_actions()
		if constructors: self.set_constructors(constructors)
		if input_data: self.add_input_data(input_data)
		self.valid_filter_field_types = ["date", "year", "datetime"]

	def merge(self, order_and_key=None, overwrite_working=False):
		"""
		Merge input data on a key column.

		Parameters
		----------
		order_and_key: list
			List of dictionaries specifying `input_data` order and key for merge. Can also use
			`order_and_key_input_data` directly. Each dict in the list has
			`{id: input_data id, key: column_name for merge}`
		overwrite_working: bool
			Permission to overwrite existing working data

		TO_DO
		-----
		While `merge` validates column uniqueness prior to merge, if the column is not unique there
		is nothing the user can do about it (without manually going and refixing the input data).
		Some sort of uniqueness fix required (probably using the filters).
		"""
		if order_and_key and isinstance(order_and_key, list):
			self.order_and_key_input_data(*order_and_key)
		if self._status in STATUS_CODES.keys() - ["READY_MERGE", "READY_STRUCTURE", "READY_CATEGORY",
												  "READY_FILTER", "READY_TRANSFORM", "PROCESS_COMPLETE",
												  "MERGE_ERROR"]:
			e = "Current status: `{}` - performing `merge` is not permitted.".format(self.status)
			raise PermissionError(e)
		self.validate_merge_data
		if self.schema_settings.get("working_data",{}).get("checksum") and not overwrite_working:
			e = "Permission required to overwrite `working_data`. Set `overwrite_working` to `True`."
			raise PermissionError(e)
		# Perform merge
		df = self.perform_merge
		# Save the file to the working directory
		if "working_data" in self.schema_settings:
			del self.schema_settings["working_data"]
		self.schema_settings["working_data"] = self.save_data(df, prefix="working")

		# Review for any existing actions
		# self.validate_actions

		self._status = "READY_STRUCTURE"

	def structure(self, name):
		"""
		Return a 'markdown' version of the formal structure for a specific `name` field.

		Returns
		-------
		list
			(nested) strings in structure format
		"""
		markdown = self.field(name).get("structure")
		if markdown:
			return self.build_structure_markdown(markdown)
		return []

	def set_structure(self, **kwargs):
		"""
		Receive a list of methods of the form::

			{
				"schema_field1": ["action", "column_name1", ["action", "column_name2"]],
				"schema_field2": ["action", "column_name1", "modifier", ["action", "column_name2"]],
			}

		The format for defining a `structure` is as follows::

			[action, column_name, [action, column_name]]

		e.g.::

			["CATEGORISE", "+", ["ORDER", "column_1", "column_2"]]

		This permits the creation of quite expressive wrangling structures from simple building
		blocks.

		Every task structure must start with an action to describe what to do with the following
		terms. There are several "actions" which can be performed, and some require action
		modifiers:

			* NEW: Add in a new column, and populate it according to the value in the "new" constraint

			* RENAME: If only 1 item in list of source fields, then rename that field

			* ORDER: If > 1 item in list of source fields, pick the value from the column, replacing each value with one from the next in the order of the provided fields

			* ORDER_NEW: As in ORDER, but replacing each value with one associated with a newer "dateorder" constraint

				* MODIFIER: `+` between terms for source and source_date

			* ORDER_OLD: As in ORDER, but replacing each value with one associated with an older "dateorder" constraint

				* MODIFIER: `+` between terms for source and source_date

			* CALCULATE: Only if of "type" = "float64" (or which can be forced to float64)

				* MODIFIER: `+` or `-` before each term to define whether add or subtract

			* JOIN: Only if of "type" = "object", join text with " ".join()

			* CATEGORISE: Only if of "type" = "string"; look for associated constraint, "categorise" where `True` = keep a list of categories, `False` = set True if terms found in list

				* MODIFIER:

					* `+` before terms where column values to be classified as unique

					* `-` before terms where column values are treated as boolean

		Paramaters
		----------
		kwargs: dict
			Where key is schema target field and value is list defining the structure action
		"""
		if self._status in STATUS_CODES.keys() - ["READY_STRUCTURE", "READY_CATEGORY", "READY_FILTER",
												  "READY_TRANSFORM", "PROCESS_COMPLETE", "STRUCTURE_ERROR"]:
			e = "Current status: `{}` - performing `set_structure` not permitted.".format(self.status)
			raise PermissionError(e)
		self.validate_merge
		category_check = False
		for field_name in kwargs:
			if field_name not in self.all_field_names:
				e = "Term `{}` not a valid field for this schema.".format(field_name)
				raise ValueError(e)
			schema_field = self.field(field_name)
			schema_field["structure"] = self.set_field_structure(kwargs[field_name])
			# Set unique field structure categories
			has_category = []
			for term in set(self.flatten_category_fields(kwargs[field_name])):
				term = term.split("::")
				modifier = term[0]
				# Just in case
				column = "::".join(term[1:])
				has_category.append(self.set_field_structure_categories(modifier, column))
			if has_category:
				category_check = True
				if (not schema_field.get("constraints", {}).get("category") and
					schema_field["type"] == "boolean"):
					if not schema_field.get("constraints"):
						schema_field["constraints"] = {}
					# Boolean fields have an implied constraint
					schema_field["constraints"]["category"] = [
						{
							"name": True
						},
						{
							"name": False
						}
					]
				if not schema_field.get("constraints", {}).get("category"):
					e = "Field `{}` has no `category` constraints. Please `set_field_category`."
					raise ValueError(e.format(field_name))
				schema_field["constraints"]["category_input"] = has_category
			# Validation would not add in the new values
			self.set_field(validate=False, **schema_field)
		self._status = "READY_CATEGORY"
		if not category_check: self._status = "READY_FILTER"

	def category(self, name):
		"""
		Return a 'markdown' version of assigned and unassigned category inputs for a named field
		of the form::

			{
				"categories": ["category_1", "category_2"]
				"assigned": {
					"category_1": ["term1", "term2", "term3"],
					"category_2": ["term4", "term5", "term6"]
				},
				"unassigned": ["term1", "term2", "term3"]
			}

		The format for defining a `category` term as follows::

			`term_name::column_name`

		Returns
		-------
		list of (nested) strings in structure format
		"""
		schema_field = self.field(name)
		assigned = {}
		unassigned = []
		if schema_field.get("category"):
			for marked in schema_field["category"].get("assigned", []):
				if schema_field["type"] == "boolean":
					if marked["name"]: u = "true"
					else: u = "false"
				else:
					u = marked["name"]
				assigned[u] = self.build_category_markdown(marked["category_input"])
			unassigned = self.build_category_markdown(schema_field["category"].get("unassigned", []))
		if (not schema_field.get("constraints", {}).get("category", []) or
			not schema_field.get("constraints", {}).get("category_input")):
			e = "Field `{}` has no available categorical data.".format(field_name)
			raise ValueError(e)
		if not schema_field.get("category"):
			unassigned = self.build_category_markdown(schema_field["constraints"]["category_input"])
		# Deals with boolean `True`/`False` case...
		if schema_field["type"] == "boolean":
			categories = ["true", "false"]
		else:
			categories = [c["name"] for c in schema_field["constraints"]["category"]]
		response = {
			"categories": categories,
			"assigned": assigned,
			"unassigned": unassigned
		}
		return response

	def set_category(self, **kwargs):
		"""
		Receive a list of categories of the form::

			{
				"schema_field1": {
					"category_1": ["term1", "term2", "term3"],
					"category_2": ["term4", "term5", "term6"]
				}
			}

		The format for defining a `category` term as follows::

			`term_name::column_name`
		"""
		if self._status in STATUS_CODES.keys() - ["READY_CATEGORY", "READY_FILTER",
												  "READY_TRANSFORM", "PROCESS_COMPLETE", "CATEGORY_ERROR"]:
			e = "Current status: `{}` - performing `set_category` not permitted.".format(self.status)
			raise PermissionError(e)
		self.validate_structure
		for field_name in kwargs:
			if field_name not in self.all_field_names:
				e = "Term `{}` not a valid field for this schema.".format(field_name)
				raise ValueError(e)
			schema_field = self.field(field_name)
			field_category = schema_field.get("constraints", {}).get("category_input")
			# Validation checks
			if not field_category:
				e = "Field `{}` has no available categorical data.".format(field_name)
				raise ValueError(e)
			if schema_field["type"] == "boolean":
				cat_diff = set(kwargs[field_name].keys()) - set(["true", "false"])
			else:
				field_category = schema_field.get("constraints", {}).get("category", [])
				cat_diff = set(kwargs[field_name].keys()) - set([c["name"] for c in field_category])
			if cat_diff:
				e = "Field `{}` has invalid categories `{}`.".format(field_name, cat_diff)
				raise ValueError(e)
			# Get assigned category_inputs
			assigned = []
			for name in kwargs[field_name]:
				c_name = name
				if schema_field["type"] == "boolean":
					if c_name == "true":
						c_name = True
					else:
						c_name = False
				category_term = {
					"name": c_name,
					"category_input": []
				}
				input_terms = {}
				for term in kwargs[field_name][name]:
					term = term.split("::")
					column = term[-1]
					term = "::".join(term[:-1])
					if not input_terms.get(column):
						input_terms[column] = []
					if term not in input_terms[column]:
						input_terms[column].append(term)
				# process input_terms
				for column in input_terms:
					category_term["category_input"].append(
						{
							"column": column,
							"terms": input_terms[column]
						}
					)
				# append to category
				assigned.append(category_term)
			# Get unassigned category_inputs
			unassigned = []
			for terms in schema_field["constraints"]["category_input"]:
				all_terms = terms["terms"]
				all_assigned_terms = []
				for assigned_term in assigned:
					assigned_input_terms = []
					for assigned_input in assigned_term["category_input"]:
						if assigned_input["column"] == terms["column"]:
							assigned_input_terms = assigned_input["terms"]
							break
					# validate
					if set(assigned_input_terms) - set(all_terms):
						e = "Field `{}` has invalid input category terms `{}`."
						raise ValueError(e.format(field_name,
												  set(assigned_input_terms) - set(all_terms)))
					# extend
					all_assigned_terms.extend(assigned_input_terms)
				# validate if duplicates
				if len(all_assigned_terms) > len(set(all_assigned_terms)):
					e = "Field `{}` has duplicate input category terms `{}`."
					# https://stackoverflow.com/a/9835819
					seen = set()
					dupes = [x for x in all_assigned_terms if x not in seen and not seen.add(x)]
					raise ValueError(e.format(field_name, dupes))
				unassigned_terms = list(set(all_terms) - set(all_assigned_terms))
				unassigned.append(
						{
							"column": terms["column"],
							"terms": unassigned_terms
						}
					)
			# Set the category
			schema_field["category"] = {
				"assigned": assigned,
				"unassigned": unassigned
			}
			# Update the field
			self.set_field(validate=False, **schema_field)
		self.reset_data_checksums(reset_output_only=True)
		self._status = "READY_TRANSFORM"

	def filter(self, name):
		"""
		Return the filter settings for a named field. If there are no filter settings, return None.

		Raises
		------
		TypeError if setting a filter on this field type is not permitted.

		Returns
		-------
		dict of filter settings, or None
		"""
		schema_field = self.field(field_name)
		# Validate the filter
		if schema_field["type"] not in self.valid_filter_field_types:
			e = "Filters cannot be set on field of type `{}`.".format(schema_field["type"])
			raise TypeError(e)
		return schema_field.get("filter", None)

	def set_filter(self, field_name, filter_name, filter_date=None, foreign_field=None):
		"""
		Sets the filter settings for a named field after validating all parameters.

		NOTE: filters can only be set on date-type fields. **whyqd** offers only rudimentary post-
		wrangling functionality. Filters are there to, for example, facilitate importing data
		outside the bounds of a previous import.

		This is also an optional step. By default, if no filters are present, the transformed output
		will include `ALL` data.

		Parameters
		----------
		field_name: str
			Name of field on which filters to be set
		filter_name: str
			Name of filter type from the list of valid filter names
		filter_date: str (optional)
			A date in the format specified by the field type
		foreign_field: str (optional)
			Name of field to which filter will be applied. Defaults to `field_name`

		Raises
		------
		TypeError if setting a filter on this field type is not permitted.
		ValueError for any validation failures.
		"""
		if self._status in STATUS_CODES.keys() - ["READY_FILTER", "READY_TRANSFORM",
												  "PROCESS_COMPLETE", "FILTER_ERROR"]:
			e = "Current status: `{}` - performing `set_filter` not permitted.".format(self.status)
			raise PermissionError(e)
		self.validate_category
		schema_field = self.field(field_name)
		# Validate the filter
		if schema_field["type"] not in self.valid_filter_field_types:
			e = "Filters cannot be set on field of type `{}`.".format(schema_field["type"])
			raise TypeError(e)
		filter_settings = {}
		for fset in self.default_filters["filter"]["modifiers"]:
			if fset["name"] == filter_name:
				filter_settings = fset
				break
		if not filter_settings:
			e = "Filter: `{}` is not a valid filter-type.".format(filter_name)
			raise TypeError(e)
		if filter_settings["date"]:
			if not filter_date:
				e = "Filter: `{}` requires a `{}` for filtering.".format(filter_name,
																		 schema_field["type"])
				raise ValueError(e)
			_c.check_date_format(schema_field["type"], filter_date)
		else:
			# Make sure
			filter_date = False
		# Validate the foreign field
		if foreign_field:
			if foreign_field not in self.all_field_names:
				e = "Filter foreign field `{}` is not a valid field.".format(foreign_field)
				raise ValueError(e)
		else:
			foreign_field = schema_field["name"]
		# Set the filter
		schema_field["filter"] = {
			"field": foreign_field,
			"modifiers": {
				"name": filter_settings["name"],
				"date": filter_date
			}
		}
		# Update the field
		self.set_field(validate=False, **schema_field)
		self.reset_data_checksums(reset_output_only=True)
		self._status = "READY_TRANSFORM"

	def transform(self, overwrite_output=False):
		"""
		Implement the method to transform input data into output data.

		Parameters
		----------
		overwrite_output: bool
			Permission to overwrite existing output data
		"""
		if self._status in STATUS_CODES.keys() - ["READY_TRANSFORM", "PROCESS_COMPLETE"]:
			e = "Current status: `{}` - performing `transform` is not permitted.".format(self.status)
			raise PermissionError(e)
		self.validates
		if self.schema_settings.get("output_data",{}).get("checksum") and not overwrite_output:
			e = "Permission required to overwrite `output_data`. Set `overwrite_output` to `True`."
			raise PermissionError(e)
		# Perform the transformation according to the method
		df = self.perform_transform
		self.schema_settings["process_date"] = _c.get_now()
		# Save the file to the working directory
		if "output_data" in self.schema_settings:
			del self.schema_settings["output_data"]
		self.schema_settings["output_data"] = self.save_data(df, prefix="output")
		self._status = "PROCESS_COMPLETE"

	#########################################################################################
	# CITATION
	#########################################################################################

	@property
	def citation(self):
		"""
		Present a citation and validation report for this method. If citation data has been included
		in the `constructor` then that will be included.

		A citation is a special set of fields, with options for:

		* **authors**: a list of author names in the format, and order, you wish to reference them
		* **date**: publication date (uses transformation date, if not provided)
		* **title**: a text field for the full study title
		* **repository**: the organisation, or distributor, responsible for hosting your data (and your method file)
		* **doi**: the persistent `DOI <http://www.doi.org/>`_ for your repository

		Format for citation is:

			author/s, date, title, repository, doi, hash (for output data), [input sources: URI, hash]

		Returns
		-------
		str
			Text ready for citation.
		"""
		# Validate the input and output
		self.validate_input_data
		self.validate_transform
		citation = []
		ctn = self.schema_settings.get("constructors", {}).get("citation", {})
		if ctn.get("authors"):
			if isinstance(ctn["authors"], list): citation.extend(ctn["authors"])
			else: citation.append(ctn["authors"])
		# Date
		citation.append(ctn.get("date", self.schema_settings["process_date"]))
		if ctn.get("title", self.details.get("title")):
			citation.append(ctn.get("title", self.details.get("title")))
		if ctn.get("repository"): citation.append(ctn["repository"])
		if ctn.get("doi"): citation.append(ctn["doi"])
		# Output hash
		citation.append(self.schema_settings["output_data"]["checksum"])
		# Input data
		input_reference = []
		for input_data in self.input_data:
			input_reference.append("{}, {}".format(input_data["original"], input_data["checksum"]))
		citation.append("[input sources: {}]".format("; ".join(input_reference)))
		return ", ".join(citation)

	#########################################################################################
	# SUPPORT FUNCTIONS
	#########################################################################################

	@property
	def status(self):
		return STATUS_CODES[self._status]

	def set_directory(self, directory):
		_c.check_path(directory)
		self.schema_settings["directory"] = directory
		self.directory = directory

	@property
	def constructors(self):
		"""
		Constructors are additional metadata to be included with the `method`. Ordinarily, this is
		a dictionary of key:value pairs defining any metadata that may be used post-wrangling and
		need to be maintained with the target data.
		"""
		return deepcopy(self.schema_settings.get("constructors"))

	def set_constructors(self, constructors, overwrite=False):
		"""
		Define additional metadata to be included with the `method`.

		Citation data must be specifically included as:

			{
				"citation": {
					"authors": ["Author Name 1", "Author Name 2"],
					"title": "Citation Title",
					"repository": "Data distributor",
					"doi": "Persistent URI"
				}
			}

		Parameters
		----------
		constructors: dict
			A set of key:value pairs. These will not be validated, or used during transformation.
		overwrite: boolean
			To overwrite any existing data in the constructor, set to True

		Raises
		------
		TypeError if not a dict.
		"""
		if not constructors or not isinstance(constructors, dict):
			e = "Method constructor is not a valid dict."
			raise TypeError(e)
		if self.schema_settings.get("constructors") and not overwrite:
			self.schema_settings["constructors"] = {**self.schema_settings["constructors"],
													**deepcopy(constructors)}
		else:
			self.schema_settings["constructors"] = deepcopy(constructors)

	def reset_data_checksums(self, reset_status=False, reset_output_only=False):
		"""
		If input or working data are modified, then the checksums for working and output data
		must be deleted (i.e. they're no longer valid and everything else must be re-run).

		Parameters
		----------
		reset_status: bool
			Requires a deliberate choice. Default False.
		reset_output_only: bool
			Requires a deliberate choice. Default False. Only resets output data.
		"""
		if (not (reset_status or reset_output_only) and
			(self.schema_settings.get("output_data", {}).get("checksum") or
			 self.schema_settings.get("working_data", {}).get("checksum"))):
			e = "Permission required to reset data. Set `reset_status` or `reset_output_only` to `True`."
			raise PermissionError(e)
		if reset_status and self.schema_settings.get("working_data", {}).get("checksum"):
			del self.schema_settings["working_data"]["checksum"]
		if ((reset_status or reset_output_only) and
			self.schema_settings.get("output_data", {}).get("checksum")):
			del self.schema_settings["output_data"]["checksum"]

	#########################################################################################
	# CREATE & MODIFY INPUT DATA
	#########################################################################################

	@property
	def input_data(self):
		return deepcopy(self.schema_settings.get("input_data", []))

	def pretty_print_input_data(self, format="rst"):
		# https://github.com/astanin/python-tabulate#table-format
		response = ""
		for data in self.input_data:
			_id = data["id"]
			_df = pd.DataFrame(data["dataframe"])
			_df = tabulate(_df, headers="keys", tablefmt=format)
			response += HELP_RESPONSE["data"].format(_id, _df)
		return response

	def add_input_data(self, input_data, reset_status=False):
		"""
		Provide a list of strings, each the filename of input data for wrangling.

		Parameters
		----------
		input_data: str or list of str
			Each input data can be a filename, or a file_source (where filename is remote)
		reset_status: bool
			Requires a deliberate choice. Default False.

		Raises
		------
		TypeError if not a list of str.
		"""
		valid = False
		if isinstance(input_data, str):
			input_data = [input_data]
		if isinstance(input_data, list):
			valid = all([isinstance(i, str) for i in input_data])
		if not valid:
			e = "`{}` is not a valid list of input data.".format(input_data)
			raise TypeError(e)
		self.schema_settings["input_data"] = self.schema_settings.get("input_data", [])
		self.reset_data_checksums(reset_status=reset_status)
		for file_source in input_data:
			# Check if the filename is remote
			file_root = "/".join(file_source.split("/")[:-1])
			source = self.directory + file_source.split("/")[-1]
			if _c.check_uri(file_source):
				# File at remote URI
				urllib.request.urlretrieve(file_source, source)
			elif file_root:
				try:
					# File in another directory
					copyfile(file_source, source)
				except SameFileError:
					pass
			_id = str(uuid.uuid4())
			summary_data = _c.get_dataframe_summary(source)
			data = {
				"id": _id,
				"checksum": _c.get_checksum(source),
				"file": _c.rename_file(source, _id),
				"original": file_source,
				"dataframe": summary_data["df"],
				"columns": summary_data["columns"]
			}
			self.schema_settings["input_data"].append(data)
		if input_data and self.input_data:
			self._status = "READY_MERGE"

	def remove_input_data(self, _id, reset_status=False):
		"""
		Remove an input data source defined by a source _id. If data have already been merged,
		reset data processing, or raise an error.

		Parameters
		----------
		_id: str
			Unique id for an input data source. View all input data from `input_data`
		reset_status: bool
			Requires a deliberate choice. Default False.

		Raises
		------
		TypeError if not a list of str.
		"""
		self.reset_data_checksums(reset_status=reset_status)
		if self.schema_settings.get("input_data", []):
			self.schema_settings["input_data"] = [data for data in self.schema_settings["input_data"]
												  if data["id"] != _id]
		if not self.input_data:
			self._status = "WAITING"

	#########################################################################################
	# MERGE HELPERS
	#########################################################################################

	@property
	def perform_merge(self):
		"""
		Helper function to perform the merge. Also used by merge_validation step.

		Returns
		-------
		DataFrame
			Merged dataframe derived from input_data
		"""
		# Pandas 1.0 says `dtype = "string"` is possible, but it isn't currently working
		# defaulting to `dtype = object` ...
		# Note, this is done to avoid any random column processing
		df = _c.get_dataframe(self.directory + self.schema_settings["input_data"][0]["file"],
							  dtype = object)
		df_key = self.schema_settings["input_data"][0]["key"]
		missing_keys = []
		for data in self.schema_settings["input_data"][1:]:
			# defaulting to `dtype = object` ...
			dfm = _c.get_dataframe(self.directory + data["file"], dtype = object)
			dfm_key = data["key"]
			missing_keys.append(dfm_key)
			df = pd.merge(df, dfm, how="outer",
						  left_on=df_key, right_on=dfm_key,
						  indicator=False)
		# Where left key values null, copy any values in the right join-field (i.e. no key match)
		keys = [m for m in df.columns if any(k == m[:len(k)] for k in missing_keys)]
		for key in keys:
			df.loc[:, df_key] = np.where(pd.isnull(df[df_key]),
										 df[key],
										 df[df_key])
		# Deduplicate any columns after merge (and deduplicate the deduplicate in case of artifacts)
		df.columns = self.deduplicate_columns(self.deduplicate_columns(df.columns))
		return df

	def order_and_key_input_data(self, *order_and_key):
		"""
		Reorder a list of input_data prior to merging, and add in the merge keys.

		Parameters
		----------
		order_and_key: list of dicts
			Each dict in the list has {id: input_data id, key: column_name for merge}

		Raises
		------
		ValueError not all input_data are assigned an order and key.
		"""
		self.validate_input_data
		reordered_data = []
		for oak in order_and_key:
			for data in self.input_data:
				columns = [c["name"] for c in data["columns"]]
				if oak["id"] == data["id"] and oak["key"] in columns:
					data["key"] = oak["key"]
					reordered_data.append(data)
		if len(reordered_data) != len(self.schema_settings["input_data"]):
			e = "List mismatch. Input-data different from list submitted for ordering."
			raise ValueError(e)
		self.schema_settings["input_data"] = reordered_data
		self._status = "READY_MERGE"

	def deduplicate_columns(self, idx, fmt=None, ignoreFirst=True):
		"""
		Source: https://stackoverflow.com/a/55405151
		Returns a new column list permitting deduplication of dataframes which may result from merge.

		Parameters
		----------
		idx: df.columns (i.e. the indexed column list)
		fmt: A string format that receives two arguments: name and a counter. By default: fmt='%s.%03d'
		ignoreFirst: Disable/enable postfixing of first element.

		Returns
		-------
		list of strings
			Updated column names
		"""
		idx = pd.Series(idx)
		duplicates = idx[idx.duplicated()].unique()
		fmt = '%s_%03d' if fmt is None else fmt
		for name in duplicates:
			dups = idx==name
			ret = [fmt%(name,i) if (i!=0 or not ignoreFirst) else name
				   for i in range(dups.sum())]
			idx.loc[dups] = ret
		# Fix any fields with the same name as any of the target fields
		for name in self.all_field_names:
			dups = idx==name
			ret = ["{}__dd".format(name) for i in range(dups.sum())]
			idx.loc[dups] = ret
		return pd.Index(idx)

	@property
	def working_column_list(self):
		if self.schema_settings.get("working_data"):
			return [c["name"] for c in self.schema_settings["working_data"]["columns"]]
		return []

	def working_data_field(self, column):
		if self.schema_settings.get("working_data"):
			for field in self.schema_settings["working_data"]["columns"]:
				if field["name"] == column:
					return field
		e = "Field `{}` is not in the working data.".format(column)
		raise KeyError(e)

	@property
	def working_dataframe(self):
		if "working_data" in self.schema_settings:
			source = self.directory + self.schema_settings["working_data"]["file"]
			return _c.get_dataframe(source, dtype=object)
		e = "No working data found."
		raise ValueError(e)

	@property
	def working_data(self):
		return deepcopy(self.schema_settings.get("working_data", {}))

	#########################################################################################
	# STRUCTURE HELPERS
	#########################################################################################

	def flatten_category_fields(self, structure, modifier=None):
		modifier_list = ["+", "-"]
		response = []
		for sublist in structure:
			if isinstance(sublist, list):
				response.extend(get_category_fields(sublist))
		if structure[0] == "CATEGORISE":
			for strut in _c.chunks(structure[1:], len(modifier_list)):
				if len(strut) < len(modifier_list):
					continue
				if strut[0] in modifier_list:
					if isinstance(strut[1], list):
						response.extend(get_category_fields(strut[1], modifier=strut[0]))
					if isinstance(strut[1], str):
						response.append(strut[0] + "::" + strut[1])
		if structure[0] != "CATEGORISE" and modifier is not None:
			for strut in structure[1:]:
				if isinstance(strut, str):
					response.append(modifier + "::" + strut)
				if isinstance(strut, list):
					response.extend(get_category_fields(strut, modifier=modifier))
		return response

	def set_field_structure_categories(self, modifier, column):
		"""
		If a structure `action` is `CATEGORISE`, then specify the terms available for
		categorisation. Each field must have a modifier, including the first (e.g. +A -B +C).

		The `modifier` is one of:

			- `-`: presence/absence of column values as true/false for a specific term
			- `+`:  unique terms in the field must be matched to the unique terms defined by the
			`field` `constraints`

		As with `set_structure`, the recursive step of managing nested structures is left to the
		calling function.

		Parameters
		----------
		modifier: str
			One of `-` or `+`
		column: str
			Must be a valid column from `working_column_list`
		"""
		if column not in self.working_column_list:
			return []
		category_list = [True, False]
		# Get the modifier: + for uniqueness, and - for boolean treatment
		if modifier == "+":
			category_list = list(self.working_dataframe[column].dropna().unique())
		structure_categories = {
			"terms": category_list,
			"column": column
		}
		return structure_categories

	def set_field_structure(self, structure_list):
		"""
		A recursive function which traverses a list defined by `*structure`, ensuring that the first
		term is an `action`, and that the subsequent terms conform to that action's requirements.
		Nested structures are permitted.

		The format for defining a `structure` is as follows::

			[action, column_name, [action, column_name]]

		e.g.::

			["CATEGORISE", "+", ["ORDER", "column_1", "column_2"]]

		This permits the creation of quite expressive wrangling structures from simple building
		blocks.

		Parameters
		----------
		structure_list: list
			Each structure list must start with an `action`, with subsequent terms conforming to the
			requirements for that action. Nested actions defined by nested lists.
		"""
		# Sets or updates a structure, and sets any required categories
		structure = []
		for i, term in enumerate(structure_list):
			if i == 0:
				# Validate the rest of the structure_list first
				action = self.default_actions[term]
				if not action.has_valid_structure(self.working_column_list, structure_list[1:]):
					e = "Task action `{}` has invalid structure `{}`.".format(term, structure_list)
					raise ValueError(e)
				structure.append(action.settings)
				continue
			if isinstance(term, list):
				# Deal with nested structures
				structure.append(self.set_field_structure(term))
				continue
			if action.name == "NEW":
				# Special case for "NEW" action
				new_term = {
					"value": term,
					"type": _c.get_field_type(term)
				}
				structure.append(new_term)
				continue
			if term in action.modifier_names:
				structure.append(action.get_modifier(term))
				continue
			if term in self.working_column_list:
				structure.append(self.working_data_field(term))
				continue
		return structure

	@property
	def default_action_types(self):
		"""
		Default list of actions available to define methods. Returns only a list
		of types. Details for individual default actions can be returned with
		`default_action_settings`.

		Returns
		-------
		list
		"""
		return list(self.default_actions.keys())

	def default_action_settings(self, action):
		"""
		Get the default settings available for a specific action type.

		Parameters
		----------
		action: string
			A specific term for an action type (as listed in `default_action_types`).

		Returns
		-------
		dict, or empty dict if no such `action_type`
		"""
		if action in self.default_actions:
			return deepcopy(self.default_action[action].settings)
		return {}

	def build_default_actions(self):
		"""
		Build the default actions for presentation to the user as options.

		Returns
		-------
		dict of Actions
		"""
		default_actions = {action["name"]: Action(validate=False, **action)
						   for action in _c.get_settings("actions")["fields"]}
		return default_actions

	def build_structure_markdown(self, structure):
		"""
		Recursive function that iteratively builds a markdown version of a formal field structure.

		Returns
		-------
		list
		"""
		markdown = []
		for strut in structure:
			if isinstance(strut, list):
				markdown.append(self.build_structure_markdown(strut))
				continue
			if strut.get("name"):
				markdown.append(strut["name"])
				continue
			if strut.get("value"):
				markdown.append(strut["value"])
		return markdown

	#########################################################################################
	# CATEGORY HELPERS
	#########################################################################################

	def build_category_markdown(self, category_input):
		"""
		Converts category_terms dict into a markdown format::

			["term1", "term2", "term3"]

		Where the format for defining a `category` term as follows::

			`term_name::column_name`

		From::

			{
				"column": "column_name",
				"terms": [
					"term_1",
					"term_2",
					"term_3"
				]
			}
		"""
		markdown = []
		for column in category_input:
			for term in column["terms"]:
				markdown.append("{}::{}".format(term, column["column"]))
		return markdown

	#########################################################################################
	# TRANSFORM HELPERS
	#########################################################################################

	@property
	def perform_transform(self):
		"""
		Helper function to perform the transformation. Also used by validate_transform step.

		Returns
		-------
		DataFrame
			Transformed dataframe derived from method
		"""
		# Keep the original merge key field as a dtype=object to avoid messing with text formatting
		# e.g. leading 0s in a reference id
		set_dtypes = {}
		set_dtypes[self.schema_settings["input_data"][0]["key"]] = object
		df = _c.get_dataframe(self.directory + self.schema_settings["working_data"]["file"],
							  dtype=set_dtypes)
		# Begin transformations + keep track if any need filters
		filter_list = []
		for field_name in self.all_field_names:
			field = self.field(field_name)
			if field.get("filter"):
				filter_list.append(field_name)
			df = task.perform_transform(df, field_name, field["type"], field["structure"],
										category = field.get("category", {}).get("assigned"))
		# Identify required output fields not in df, and set blank fields for these
		blank_fields = list(set(self.all_field_names) - set(df.columns))
		for blank in blank_fields:
			df[blank] = ""
		# Conclude transformation
		df = df[self.all_field_names]
		df = df.loc[df.astype(str).drop_duplicates().index]
		# Perform filter requirements - note, the order may be important, but the user should be
		# careful with brute-force filters and should only set one. If they didn't ...
		for field_name in filter_list:
			field = self.field(field_name)
			filter = field["filter"]
			if filter["modifiers"]["name"] == "LATEST":
				# https://pandas.pydata.org/pandas-docs/stable/groupby.html#splitting-an-object-into-groups
				df = df.sort_values(by=field_name)
				df = df.groupby(filter["field"])
				# Select only the latest date
				df = df.last()
				# https://stackoverflow.com/a/20461206
				df.reset_index(level=df.index.names, inplace=True)
			if filter["modifiers"]["name"] == "AFTER" and filter["modifiers"].get("date"):
				df = df[df[field_name] > filter["modifiers"]["date"]]
			if filter["modifiers"]["name"] == "BEFORE" and filter["modifiers"].get("date"):
				df = df[df[field_name] < filter["modifiers"]["date"]]
			try:
				if filter["modifiers"]["name"] in ["LATEST", "AFTER"]:
					field["filter"]["date"] = df[pd.notnull(df[field_name])][field_name].max()
				else:
					field["filter"]["date"] = df[pd.notnull(df[field_name])][field_name].min()
				self.set_field(validate=False, **field)
			except TypeError:
				# They're not actually dates but are text for some reason
				pass
		return df

	#########################################################################################
	# VALIDATE, BUILD AND SAVE
	#########################################################################################

	@property
	def validate_input_data(self):
		"""
		Test input data for checksum errors.

		Raises
		------
		ValueError on checksum failure.
		"""
		if not self.schema_settings.get("input_data", []):
			e = "Empty list. No valid input data."
			raise ValueError(e)
		for data in self.schema_settings["input_data"]:
			source = self.directory + data["file"]
			if _c.get_checksum(source) != data["checksum"]:
				e = "Checksum error on input data `{}`".format(data["original"])
				raise ValueError(e)
		return True

	@property
	def validate_merge_data(self):
		"""
		Test input data ready to merge; that it has a merge key, and that the data in that column
		are unique.

		Raises
		------
		ValueError on uniqueness failure.
		"""
		for data in self.input_data:
			if not data.get("key"):
				e = "Missing merge key on input data `{}`".format(data["original"])
				raise ValueError(e)
			source = self.directory + data["file"]
			_c.check_column_unique(source, data["key"])
		return True

	@property
	def validate_merge(self):
		"""
		Validate merge output.

		Raises
		------
		ValueError on checksum failure.

		Returns
		-------
		bool: True for validates
		"""
		df = self.perform_merge
		# Save temporary file ... has to save & load for checksum validation
		_id = str(uuid.uuid4())
		filetype = self.schema_settings["working_data"]["file"].split(".")[-1]
		filename = "".join([_id, ".", filetype])
		source = self.directory + filename
		if filetype == "csv":
			df.to_csv(source, index=False)
		if filetype == "xlsx":
			df.to_excel(source, index=False)
		checksum = _c.get_checksum(source)
		_c.delete_file(source)
		if (self.schema_settings["working_data"]["checksum"] != checksum):
			e = "Merge validation checksum failure {} != {}".format(
				self.schema_settings["working_data"]["checksum"],
				checksum)
			raise ValueError(e)
		return True

	@property
	def validate_structure(self):
		"""
		Method validates structure formats.

		Raises
		------
		ValueError on structure failure.

		Returns
		-------
		bool: True for validates
		"""
		for field_name in self.all_field_names:
			# Test structure
			structure = self.field(field_name).get("structure")
			if not structure:
				e = "Structure: {}".format(field_name)
				raise ValueError(e)
			test_structure = self.build_structure_markdown(structure)
			if structure != self.set_field_structure(test_structure):
				e = "Structure for Field `{}` is not valid".format(field_name)
				raise ValueError(e)
			# Test structure categories
			category = self.field(field_name).get("constraints", {}).get("category_input")
			if category and self.field(field_name).get("constraints", {}).get("category"):
				# Needs to have both a set of terms derived from the columns, and a
				# defined set of terms to be categorised as...
				test_category = []
				for term in set(self.flatten_category_fields(test_structure)):
					term = term.split("::")
					modifier = term[0]
					# Just in case
					column = "::".join(term[1:])
					test_category.append(self.set_field_structure_categories(modifier, column))
				if (sorted(category, key=itemgetter("column")) !=
					sorted(test_category, key=itemgetter("column"))):
					# Equality test on list of dicts requires them to be in the same order
					e = "Category for Field `{}` is not valid".format(field_name)
					raise ValueError(e)
		return True

	@property
	def validate_category(self):
		"""
		Method validates category terms.

		Raises
		------
		ValueError on category failure.

		Returns
		-------
		bool: True for validates
		"""
		for field_name in self.all_field_names:
			schema_field = self.field(field_name)
			if schema_field.get("category"):
				# Test if the category and category_input constraints are valid
				# assigned is of the form: category: [term::column]
				# constraints of the form: category -> name | category_input -> column | terms
				category = self.category(field_name).get("assigned")
				if schema_field["type"] == "boolean":
					category_constraints = ["true", "false"]
				else:
					category_constraints = [c["name"] for c in
											schema_field["constraints"].get("category", {})]
				diff = set(category.keys()) - set(category_constraints)
				if diff:
					e = "Category for Field `{}` has invalid constraints `{}`"
					raise ValueError(e.format(field_name, diff))
				category_input = {}
				for terms in category.values():
					for term in terms:
						term = term.split("::")
						if not category_input.get(term[-1]):
							category_input[term[-1]] = []
						category_input[term[-1]].append("::".join(term[:-1]))
				category_input_columns = [c["column"] for c in
										  schema_field["constraints"].get("category_input", {})]
				diff = set(category_input.keys()) - set(category_input_columns)
				if diff:
					e = "Category for Field `{}` has invalid data columns `{}`"
					raise ValueError(e.format(field_name, diff))
				category_input_terms = [c["terms"] for c in
										schema_field["constraints"].get("category_input", {})]
				category_input_terms = [item for sublist in category_input_terms
										for item in sublist]
				category_terms = [item for sublist in category_input.values()
								  for item in sublist]
				diff = set(category_terms) - set(category_input_terms)
				if diff:
					e = "Category for Field `{}` has invalid input category terms `{}`"
					raise ValueError(e.format(field_name, diff))
				diff = len(category_terms) - len(set(category_terms))
				if diff:
					e = "Category for Field `{}` has duplicate input category terms"
					raise ValueError(e.format(field_name))
		return True

	@property
	def validate_filter(self):
		"""
		Method validates filter terms.

		Raises
		------
		ValueError on filter failure.

		Returns
		-------
		bool: True for validates
		"""
		for field_name in self.all_field_names:
			schema_field = self.field(field_name)
			if not schema_field.get("filter"):
				continue
			# Filter permitted
			if schema_field["type"] not in self.valid_filter_field_types:
				e = "Filter of field `{}` is not permitted.".format(schema_field["name"])
				raise PermissionError(e)
			# Foreign key valid
			if schema_field["filter"]["field"] not in self.all_field_names:
				e = "Filter foreign key `{}` not a valid field for this schema"
				raise ValueError(e.format(schema_field["filter"]["field"]))
			# Filter type valid
			default_filters = [f["name"] for f in self.default_filters["filter"]["modifiers"]]
			if schema_field["filter"]["modifiers"]["name"] not in default_filters:
				e = "Filter: `{}` is not a valid filter-type."
				raise TypeError(e.format(schema_field["filter"]["modifiers"]["name"]))
			# Filter date valid
			if schema_field["filter"]["modifiers"]["date"]:
				_c.check_date_format(schema_field["type"],
									 schema_field["filter"]["modifiers"]["date"])
		return True

	@property
	def validate_transform(self):
		"""
		Validate output data.

		Raises
		------
		ValueError on checksum failure.

		Returns
		-------
		bool: True for validates
		"""
		df = self.perform_transform
		# Save temporary file ... has to save & load for checksum validation
		_id = str(uuid.uuid4())
		filetype = self.schema_settings["output_data"]["file"].split(".")[-1]
		filename = "".join([_id, ".", filetype])
		source = self.directory + filename
		if filetype == "csv":
			df.to_csv(source, index=False)
		if filetype == "xlsx":
			df.to_excel(source, index=False)
		checksum = _c.get_checksum(source)
		_c.delete_file(source)
		if (self.schema_settings["output_data"]["checksum"] != checksum):
			e = "Transformation validation checksum failure {} != {}".format(
				self.schema_settings["output_data"]["checksum"],
				checksum)
			raise ValueError(e)
		return True

	@property
	def validates(self):
		"""
		Method validates all steps. Sets `READY_TRANSFORM` if all pass.

		Returns
		-------
		bool: True for validates
		"""
		super().validates
		self.validate_input_data
		self.validate_merge_data
		self.validate_merge
		self.validate_structure
		self.validate_category
		self.validate_filter
		self._status = "READY_TRANSFORM"
		return True

	def build_action(self, **action):
		"""
		For a list of actions, defined as dictionaries, create and return Action objects.

		Parameters
		----------
		field: dictionary of Action parameters

		Raises
		------
		ValueError: if action fails validation

		Returns
		-------
		Action
		"""
		if not isinstance(action, dict) or not all(key in action for key in self.required_field_terms):
			e = "Action is not a valid dictionary"
			raise ValueError(e)
		action = Action(**action)
		if not action.validates:
			e = "Action `{}` of type `{}` does not validate".format(_name, _type)
			raise ValueError(e)
		return action

	def build(self):
		"""
		Build and validate the Method. Note, this replaces the Schema base-class.
		"""
		self.schema_settings["fields"] = [self.build_field(validate=False, **field) for field in
										  self.schema_settings.get("fields", [])]
		if not self.directory:
			e = "Action is not a valid dictionary"
			raise ValueError(e)
		#self.set_directory(self.directory)
		self._status = self.schema_settings.get("status", self._status)

	def save_data(self, df, filetype="xlsx", prefix=None):
		"""
		Generate a unique filename for a dataframe, save to the working directory, and return the
		unique filename and data summary.

		Parameters
		----------
		df: Pandas DataFrame
		filetype: Save the dataframe as a particular type. Default is "CSV".

		Returns
		-------
		dict
			Keys include: id, filename, checksum, df header, columns
		"""
		if df.empty:
			e = "Cannot save empty DataFrame."
			raise ValueError(e)
		if filetype not in ["xlsx", "csv"]:
			e = "`{}` not supported for saving DataFrame".format(filetype)
			raise TypeError(e)
		_id = str(uuid.uuid4())
		if prefix: _id = "_".join([prefix, _id])
		filename = "".join([_id, ".", filetype])
		source = self.directory + filename
		if filetype == "csv":
			df.to_csv(source, index=False)
		if filetype == "xlsx":
			df.to_excel(source, index=False)
		summary_data = _c.get_dataframe_summary(source)
		data = {
			"id": _id,
			"checksum": _c.get_checksum(source),
			"file": filename,
			"dataframe": summary_data["df"],
			"columns": summary_data["columns"]
		}
		return data

	def save(self, directory=None, filename=None, overwrite=False, created_by=None):
		if not directory: directory = self.directory
		self.schema_settings["status"] = self._status
		super().save(directory=directory, filename=filename,
					 overwrite=overwrite, created_by=created_by)

	#########################################################################################
	# HELP
	#########################################################################################

	def help(self, option=None):
		"""
		Get generic help, or help on a specific method.

		Paramater
		---------
		option: str
			Any of None, 'status', 'merge', 'structure', 'category', 'filter', 'transform', 'error'.

		Returns
		-------
		Help
		"""
		response = ""
		if not option or option not in ["status", "merge", "structure", "category",
										  "filter", "transform", "error"]:
			response = HELP_RESPONSE["default"].format(self.status)
		elif option != "status":
			response = HELP_RESPONSE[option]
			if option == "merge":
				for data in self.input_data:
					_id = data["id"]
					_source = data["original"]
					_df = pd.DataFrame(data["dataframe"])
					_df = tabulate(_df, headers="keys", tablefmt="rst")
					response += HELP_RESPONSE["data"].format(_id, _source, _df)
			if option == "structure":
				response = response.format(self.all_field_names,
										   self.default_action_types,
										   self.working_column_list)
				if "working_data" in self.schema_settings:
					_id = self.schema_settings["working_data"]["id"]
					_source = "method.input_data"
					_df = pd.DataFrame(self.schema_settings["working_data"]["dataframe"])
					_df = tabulate(_df, headers="keys", tablefmt="rst")
					response += HELP_RESPONSE["data"].format(_id, _source, _df)
			if option == "category":
				category_fields = []
				for field_name in self.all_field_names:
					if self.field(field_name).get("constraints", {}).get("category"):
						category_fields.append(field_name)
				response = response.format(category_fields)
			if option == "filter":
				filter_fields = []
				for field_name in self.all_field_names:
					if self.field(field_name)["type"] in self.valid_filter_field_types:
						filter_fields.append(field_name)
				response = response.format(filter_fields)
		# `status` request
		response += HELP_RESPONSE["status"].format(self.status)
		return response

HELP_RESPONSE = {
	"default": """
**whyqd** provides data wrangling simplicity, complete audit transparency, and at speed.

To get help, type:

	>>> method.help(option)

Where `option` can be any of:

	status
	merge
	structure
	category
	filter
	transform

`status` will return the current method status, and your mostly likely next steps. The other options
will return methodology, and output of that option's result (if appropriate). The `error` will
present an error trace and attempt to guide you to fix the process problem.""",
	"merge": """
`merge` will join, in order from right to left, your input data on a common column.

To add input data, where `input_data` is a filename, or list of filenames:

	>>> method.add_input_data(input_data)

To remove input data, where `id` is the unique id for that input data:

	>>> method.remove_input_data(id)

Prepare an `order_and_key` list, where each dict in the list has:

	{id: input_data id, key: column_name for merge}

Run the merge by calling (and, optionally - if you need to overwrite an existing merge - setting
`overwrite_working=True`):

	>>> method.merge(order_and_key, overwrite_working=True)

To view your existing `input_data`:

	>>> method.input_data
""",
	"structure": """
`structure` is the core of the wrangling process and is the process where you define the actions
which must be performed to restructure your working data.

Create a list of methods of the form:

	{{
		"schema_field1": ["action", "column_name1", ["action", "column_name2"]],
		"schema_field2": ["action", "column_name1", "modifier", ["action", "column_name2"]],
	}}

The format for defining a `structure` is as follows::

	[action, column_name, [action, column_name]]

e.g.::

	["CATEGORISE", "+", ["ORDER", "column_1", "column_2"]]

This permits the creation of quite expressive wrangling structures from simple building
blocks.

The schema for this method consists of the following terms:

{}

The actions:

{}

The columns from your working data:

{}
""",
	"category": """
Provide a list of categories of the form::

	{{
		"schema_field1": {{
			"category_1": ["term1", "term2", "term3"],
			"category_2": ["term4", "term5", "term6"]
		}}
	}}

The format for defining a `category` term as follows::

	`term_name::column_name`

Get a list of available terms, and the categories for assignment, by calling::

	>>> method.category(field_name)

Once your data are prepared as above::

	>>> method.set_category(**category)

Field names requiring categorisation are: {}
""",
	"filter": """
Set date filters on any date-type fields. **whyqd** offers only rudimentary post-
wrangling functionality. Filters are there to, for example, facilitate importing data
outside the bounds of a previous import.

This is also an optional step. By default, if no filters are present, the transformed output
will include `ALL` data.

Parameters
----------
field_name: str
	Name of field on which filters to be set
filter_name: str
	Name of filter type from the list of valid filter names
filter_date: str (optional)
	A date in the format specified by the field type
foreign_field: str (optional)
	Name of field to which filter will be applied. Defaults to `field_name`

There are four filter_names:

	ALL: default, import all data
	LATEST: only the latest date
	BEFORE: before a specified date
	AFTER: after a specified date

BEFORE and AFTER take an optional `foreign_field` term for filtering on that column. e.g.

	>>> method.set_filter("occupation_state_date", "AFTER", "2019-09-01", "ba_ref")

Filters references in column `ba_ref` by dates in column `occupation_state_date` after `2019-09-01`.

Field names which can be filtered are: {}
""",
	"data": """

Data id: {}
Original source: {}

{}
""",
	"status": """

Current method status: `{}`"""
}


"""
Miscellaneous tools for supporting core functionality.
"""
import json
import sys, re
import copy
import hashlib
from urllib.parse import urlparse
from datetime import date, datetime, timedelta
from pathlib import Path, PurePath
import pandas as pd
import locale
try:
	locale.setlocale(locale.LC_ALL, "en_US.UTF-8")
except locale.Error:
	# Readthedocs has a problem, but difficult to replicate
	locale.setlocale(locale.LC_ALL, "")

short_rows = 3
long_rows = 5

def get_now():
	return date.isoformat(datetime.now())

def check_path(directory):
	Path(directory).mkdir(parents=True, exist_ok=True)

def check_source(source):
	if Path(source).exists():
		return True
	else:
		e = "Source at `{}` not found.".format(source)
		raise FileNotFoundError(e)

def get_path():
	return Path(__file__).resolve().parent

def check_uri(source):
	# https://stackoverflow.com/a/38020041
    try:
        result = urlparse(source)
        return all([result.scheme, result.netloc, result.path])
    except:
        return False

def rename_file(source, newname):
	p = Path(source)
	newname = "{}{}".format(newname, p.suffix)
	p.rename(Path(p.parent, newname))
	return newname

def delete_file(source):
	try:
		assert sys.version_info >= (3,8)
	except AssertionError:
		Path(source).unlink()
	else:
		Path(source).unlink(missing_ok=True)

def get_checksum(source, load_source=True):
	# https://stackoverflow.com/a/47800021
	#checksum = hashlib.blake2b()
	#with open(source, "rb") as f:
	#	for chunk in iter(lambda: f.read(4096), b""):
	#		checksum.update(chunk)
	#return checksum.hexdigest()
	if load_source:
		df = get_dataframe(source)
	else:
		df = source
	df_checksum = pd.util.hash_pandas_object(df, index=True).values
	checksum = hashlib.blake2b
	return checksum(df_checksum).hexdigest()

def chunks(l, n):
	"""Yield successive n-sized chunks from l."""
	# https://stackoverflow.com/a/976918/295606
	for i in range(0, len(l), n):
		yield l[i:i + n]

DATE_FORMATS = {
	"date": {
		"fmt": ["%Y-%m-%d"],
		"txt": ["YYYY-MM-DD"]
	},
	"datetime": {
		"fmt": ["%Y-%m-%d %H:%M:%S", "%Y-%m-%d %H:%M:%S %Z%z"],
		"txt": ["YYYY-MM-DD hh:mm:ss","YYYY-MM-DD hh:mm:ss UTC+0000"]
	},
	"year": {
		"fmt": ["%Y"],
		"txt": ["YYYY"]
	}
}

def check_date_format(date_type, date_value):
	# https://stackoverflow.com/a/37045601
	# https://www.saltycrane.com/blog/2009/05/converting-time-zones-datetime-objects-python/
	for i, fmt in enumerate(DATE_FORMATS[date_type]["fmt"]):
		try:
			if date_value == datetime.strptime(date_value, fmt).strftime(fmt):
				return True
		except ValueError:
			continue
	txt = DATE_FORMATS[date_type]["txt"]
	e = "Incorrect date format, should be: `{}`".format(txt)
	raise ValueError(e)

def parse_dates(x):
	"""
	This is the hard-won 'trust nobody', certainly not Americans, date parser.

	Yes, this will put Errors in your data. Fix your dates.
	"""
	if pd.isnull(x): return pd.NaT
	# Check if to_datetime can handle things
	if not pd.isnull(pd.to_datetime(x, errors="coerce", dayfirst=True)):
		return date.isoformat(pd.to_datetime(x, errors="coerce", dayfirst=True))
	# Manually see if coersion will work
	x = str(x).strip()[:10]
	x = re.sub(r"[\\/,\.]","-", x)
	try:
		y, m, d = x.split("-")
	except ValueError:
		return "Error: Date ({})".format(x)
	if len(y) < 4:
		# Swap the day and year positions
		# Ignore US dates - fuck 'em
		d, m, y = x.split("-")
	# Fat finger on 1999 ... not going to check for other date errors as no way to figure out
	if y[0] == "9": y = "1" + y[1:]
	x = "{}-{}-{}".format(y, m, d)
	try:
		x = datetime.strptime(x,"%Y-%m-%d")
	except ValueError:
		try:
			x = datetime.strptime(x,"%Y-%m-%d")
		except ValueError:
			if x != "":
				return "Error: Date ({})".format(x)
			return pd.NaT
	x = date.isoformat(x)
	try:
		pd.Timestamp(x)
		return x
	except pd.errors.OutOfBoundsDatetime:
		return "Error: Date ({})".format(x)

def parse_float(x):
	"""
	Regex to extract wrecked floats: https://stackoverflow.com/a/385597
	Checked against: https://regex101.com/
	"""
	try:
		return float(x)
	except ValueError:
		re_float = re.compile("""(?x)
		   ^
			  \D*     		# first, match an optional sign *and space*
			  (             # then match integers or f.p. mantissas:
				  \d+       # start out with a ...
				  (
					  \.\d* # mantissa of the form a.b or a.
				  )?        # ? takes care of integers of the form a
				 |\.\d+     # mantissa of the form .b
			  )
			  ([eE][+-]?\d+)?  # finally, optionally match an exponent
		   $""")
		try:
			x = re_float.match(x).group(1)
			x = re.sub(r"[^e0-9,-\.]","", str(x))
			return locale.atof(x)
		except (ValueError, AttributeError):
			return np.nan

###################################################################################################
### JSON, Schema and Action get and set
###################################################################################################

def load_json(source):
	"""
	Load and return a JSON file, if it exists.

	Paramaters
	----------
	source: the filename & path to open

	Raises
	------
	JSONDecoderError if not a valid json file
	FileNotFoundError if not a valid source

	Returns
	-------
	dict
	"""
	check_source(source)
	with open(source, "r") as f:
		try:
			return json.load(f)
		except json.decoder.JSONDecodeError:
			e = "File at `{}` not valid json.".format(source)
			raise json.decoder.JSONDecodeError(e)

def save_json(data, source, overwrite=False):
	"""
	Save a dictionary as a json file. Return `False` if file `source` already exists and not
	`overwrite`.

	Parameters
	----------
	data: dictionary to be saved
	source: the filename to open, including path
	overwrite: bool, True if overwrite existing file

	Returns
	-------
	bool, True if saved, False if already exists without `overwrite`
	"""
	if Path(source).exists() and not overwrite:
		e = "`{}` already exists. Set `overwrite` to `True`.".format(source)
		raise FileExistsError(e)
	with open(source, "w") as f:
		json.dump(data, f, indent=4, sort_keys=True, default=str)
	return True

JSON_SETTING_FILES = {
	"schema": "default_fields.json",
	"filter": "default_filters.json",
	"actions": "default_actions.json"
}

def get_settings(choice):
	"""
	Return the complete json settings for a json `filename` as a dict.
	"""
	filename = JSON_SETTING_FILES[choice]
	source = PurePath.joinpath(get_path(), "settings", filename)
	return load_json(source)

def get_field_settings(field_type, key="type", setting_file="schema"):
	fields = get_settings(setting_file)["fields"]
	for f in fields:
		if f[key] == field_type:
			return f
	return {}

def get_field_type(field_type):
	if isinstance(field_type, str): return "string"
	if isinstance(field_type, bool): return "boolean"
	if isinstance(field_type, int): return "integer"
	if isinstance(field_type, float): return "number"
	if isinstance(field_type, list): return "array"
	if isinstance(field_type, dict): return "object"
	return None

###################################################################################################
### Dataframe get and set
###################################################################################################

def get_dataframe(source,
				  foreign_key = None,
				  filetype = None,
				  **kwargs):
	"""
	For a given file and directory, return an initial set of rows along with a list of the columns
	and types for each column.

	Will determine filetype (CSV or XLSX) by bruteforce, but can accept a non-compulsory filetype
	kwargs.

	Parameters
	----------
	source: Source filename;
	filetype: (Optional) must be in ['csv', 'xls', 'xlsx']
	kwargs: additional keyword-arguments for Pandas;

	Returns
	-------
	Dataframe (if fails, returns empty dataframe)
	"""
	check_source(source)
	# If the dtypes have not been set, then ensure that any provided foreign_key remains untouched
	# i.e. no forcing of text to numbers
	# Pandas 1.0 says `dtype = "string"` is possible, but it isn't currently working
	# defaulting to `dtype = object` ...
	if foreign_key and "dtype" not in kwargs:
		kwargs["dtype"] = {
			foreign_key: object
		}
	# Get file delimiter
	if filetype and filetype.lower() in ["csv", "xls", "xlsx"]:
		filetype = filetype.lower()
	else:
		filetype = source.split(".")[-1].lower()
	if filetype in ["xlsx", "xls"]:
		df = pd.read_excel(source, **kwargs)
	elif filetype == "csv":
		kwargs["encoding"] = kwargs.get("encoding", "ISO-8859-1")
		kwargs["iterator"] = True
		kwargs["chunk_size"] = kwargs.get("chunk_size", 100000)
		df = pd.read_csv(source, **kwargs)
		df_chunks = []
		loop = True
		while loop:
			try:
				chunk = df.get_chunk(kwargs["chunk_size"])
				df_chunks.append(chunk)
			except StopIteration:
				loop = False
		df = pd.concat(df_chunks, ignore_index=True)
	return df

def get_dataframe_summary(source,
						  rows = short_rows,
						  foreign_key = None,
						  **kwargs):

	"""
	For a given file and directory, return an initial set of rows along with a list of the columns
	and types for each column.

	Will determine filetype (CSV or XLSX) by bruteforce, but can accept a non-compulsory filetype
	kwargs.

	Parameters
	----------
	source: Source filename;
	rows: Number of rows to return (to reduce load times), default = short_rows;
	foreign_key: Column name of field where data are not to be messed with;
	kwargs: additional keyword-arguments for Pandas;

	Returns
	-------
	dict:
		"data": { "df": df.head(rows).to_dict(), "columns": list of dicts with column names & type}
	"""
	# limit rows for dataframe
	kwargs["nrows"] = rows
	filetype = kwargs.get("filetype")
	if filetype: del kwargs["filetype"]
	# Get dataframe
	df = get_dataframe(source, foreign_key = foreign_key, filetype = filetype, **kwargs)
	# Prepare summary
	kwargs = {
		"df": df.head(rows).to_dict(),
		"columns": [
				{"name": k, "type": "number"} if v in ["float64", "int64"]
				else {"name": k, "type": "date"} if v in ["datetime64[ns]"]
				else {"name": k, "type": "string"}
			for k,v in df.dtypes.apply(lambda x: x.name).to_dict().items()]
	}
	return kwargs

def check_column_unique(source, key):
	"""
	Test a column in a dataframe to ensure all values are unique.

	Parameters
	----------
	source: Source filename
	key: Column name of field where data are to be tested for uniqueness

	Raises
	------
	ValueError if not unique

	Returns
	-------
	bool, True if unique
	"""
	df = get_dataframe(source, key)
	if len(df[key]) != len(df[key].unique()):
		import warnings
		filename = source.split("/")[-1] # Obfuscate the path
		e = "'{}' contains non-unique rows in column `{}`".format(filename, key)
		#raise ValueError(e)
		warnings.warn(e)
	return True


import sys
try:
	assert sys.version_info >= (3,7)
except AssertionError:
	from whyqd.schema.field import Field
	from whyqd.schema.schema import Schema
else:
	from .field import Field
	from .schema import Schema


import sys
try:
	assert sys.version_info >= (3,7)
except AssertionError:
	from whyqd.method.action import Action
	from whyqd.method.method import Method
else:
	from .action import Action
	from .method import Method


"""
Target Fields
-------------

Definitions and customisation for canonical destination fields forming the target schema.
"""
from copy import deepcopy

import whyqd.common as _c

class Field:
	"""
	Field definitions for a schema. A list of fields defines a schema.  Minimum arguments to create
	a field are a `name` and a `type`.
	"""
	def __init__(self, name, field_type, **kwargs):
		self.common_settings = _c.get_settings("schema")["common"]
		self.default_settings = _c.get_field_settings(field_type)
		self.default_constraints = self.default_settings["constraints"]["properties"]
		self.default_filters = _c.get_settings("filter")["filter"]["modifiers"]
		self.default_format = self.default_settings.get("format", {})
		self._name = "_".join(name.split(" ")).lower()
		self._type = field_type
		self.field_settings = {
			"name": self._name,
			"type": self._type,
			"constraints": {}
		}
		kwargs = deepcopy(kwargs)
		if kwargs:
			if kwargs.pop("validate", True):
				self.set_all(**kwargs)
			else:
				self.field_settings = {**self.field_settings, **kwargs}

	def __repr__(self):
		"""
		Returns the string representation of the model.
		"""
		return "Field: `{}` of type `{}`".format(self._name, self._type)

	@property
	def name(self):
		"""
		Field name

		Returns
		-------
			str: field name
		"""
		return self.field_settings.get("name")

	@property
	def type(self):
		"""
		Field type

		Returns
		-------
			str: field type
		"""
		return self.field_settings.get("type")

	@property
	def format(self):
		"""
		Field format

		Returns
		-------
			str: field format
		"""
		return self.field_settings.get("format", "default")

	@property
	def required(self):
		"""
		Field required

		Returns
		-------
			bool: true if required
		"""
		return self.field_settings["constraints"].get('required', False)

	@property
	def constraints(self):
		"""
		Field constraints

		Returns
		-------
			dict: dict of field constraints
		"""
		return deepcopy(self.field_settings.get('constraints', {}))

	@property
	def settings(self):
		"""
		Field settings

		Returns
		-------
			dict: settings
		"""
		response = deepcopy(self.field_settings)
		# Remove empty category fields if none used
		if not response["constraints"]: del response["constraints"]
		return response

	@property
	def validates(self):
		"""
		Field validates with minimum required keys

		Returns
		-------
			bool: True for has all keys
		"""
		if not all(key in self.settings for key in self.default_settings["required"]):
			e = "Required `field` parameters not present for `{}`.".format(self._name)
			raise ValueError(e)
		return True

	def set_constraint(self, constraint, value):
		"""
		Set field constraints. User-defined constraints will be automatically passed through, as
		long as the constraints validate.

		Raises
		------
			KeyError: if constraint is not permitted in this field
			ValueError: if value is out of range of that permitted by this constraint
		"""
		if constraint not in self.default_constraints:
			e = "`{}` not permitted in list of constraints".format(constraint)
			raise KeyError(e)
		value = deepcopy(value)
		constraint_type = _c.get_field_type(value)
		# Check for boolean constraints
		if (constraint in ["unique", "required"] and constraint_type != "boolean"):
			e = "`{}` not a permitted value type for `{}`".format(constraint_type, constraint)
			raise ValueError(e)
		# Check for appropriate minimum and maximum constraints
		if constraint in ["minimum", "maximum"]:
			if (self.default_constraints[constraint]["type"] in ["string", "integer", "number"] and
				constraint_type not in ["integer", "number"]):
				e = "Constraint `{}` should be an integer or number.".format(constraint)
				raise ValueError(e)
			if (self.default_constraints[constraint]["type"] in ["string", "integer"] and
				constraint_type == "number"):
				e = "Constraint `{}` should be an integer".format(constraint)
				raise ValueError(e)
		if constraint == "category":
			# Category constraints are very specific
			if not (isinstance(value, list) and
					isinstance(value[0], dict) and
					all(["name" in v for v in value])):
				e = "Terms defining this category constraint must be presented as a list of dicts with keys `name` & (optional) `description`"
				raise KeyError(e)
			# Each term is an object with `name`/`description`, where`name` defined by a `type`
			# Test if terms are of the required type
			terms = [t["name"] for t in value]
			# Not sure what my intention is here. I mean, names are going to be strings, right?
			# Maybe need to specify the types in the array somewhere?
			# leaving it for now...
			#required_type = self.default_constraints["category"]["terms"]["type"]
			#if not all([_c.get_field_type(t) == required_type for t in terms]):
			#	e = "`{}` are not all of type `{}`".format(terms, required_type)
			#	raise ValueError(e)
			# Test if terms are unique
			if len(terms) < len(set(terms)):
				e = "Category list `{}` has duplicate terms".format(value)
				raise ValueError(e)
		if constraint == "valueType":
			# This is to set the `type` for the contents of an array
			if self._type != "array":
				import warnings
				e = "`valueType` will be ignored for fields which are not type `array`."
				warnings.warn(e)
			if value not in self.default_constraints["valueType"]["type"]:
				e = "valueType `{}` not one of `{}`".format(value, self.default_constraints["valueType"]["type"])
				raise ValueError(e)
		if constraint == "filter":
			# This is a user-defined constraint which can be applied to any date-related field type
			if not value.get("modifiers"):
				e = "`{}` has missing `modifiers` in `filter`".format(constraint)
				raise KeyError(e)
			if "field" in value and not (isinstance(value["field"], str) or
										 isinstance(value["field"], bool)):
				e = "Filter field `{}` is not of type `string` or `True`".format(value["field"])
				raise KeyError(e)
			if value.get("field", "foreignKey") == "foreignKey":
				value["field"] = True
		self.field_settings["constraints"][constraint] = value

	def set_missing(self, value):
		"""
		Set default value for missing values.

		Raises
		------
			ValueError: if value is out of range of that permitted by this field
		"""
		value = deepcopy(value)
		_type = _c.get_field_type(value)
		if value != "" or _type != self.field_settings["type"]:
			e = "`{}` not a permitted value type for `{}`".format(value, self.field_settings["type"])
			raise ValueError(e)
		self.field_settings["missing"] = value

	def set_format(self, value="default"):
		"""
		Set field format.

		Raises
		------
			ValueError: if value is not permitted for this field
		"""
		value = deepcopy(value)
		_type = _c.get_field_type(value)
		if value not in self.default_format.get("category", ["default"]):
			e = "`{}` not a permitted format".format(value)
			raise ValueError(e)
		self.field_settings["format"] = value

	def set_foreign_key(self, value):
		"""
		Set foreign_key value.

		Raises
		------
			TypeError: if value is not a bool
		"""
		value = deepcopy(value)
		if not isinstance(value, bool):
			e = "`{}` is not boolean".format(value)
			raise TypeError(e)
		if value:
			self.field_settings["foreignKey"] = True
		elif "foreignKey" in self.field_settings:
			del self.field_settings["foreignKey"]

	def set_all(self, **kwargs):
		"""
		Set all keys and values for this field. Note, this also permits updating of existing
		`name`, but not `type` (which would cause a disaster).

		Arguments
		---------
			kwargs: dict of all the permitted key: value pairs for this field

		Raises
		------
			ValueError: if `name` contains spaces
			TypeError: if string terms are not strings
		"""
		kwargs = deepcopy(kwargs)
		if "type" in kwargs and kwargs["type"] != self._type:
			e = "Warning: You are not permitted to change an existing Field type. Better to delete & start again."
			print(e)
		for key, value in kwargs.items():
			if key == "type":
				# Ignore any attempts to change `type`
				continue
			if key == "foreignKey":
				self.set_foreign_key(value)
				continue
			if key in self.common_settings:
				if not isinstance(value, str):
					e = "`{}` is not a string".format(value)
					raise TypeError(e)
				if key == "name":
					value = "_".join(value.split(" ")).lower()
				self.field_settings[key] = value
				continue
			if key == "format":
				self.set_format(value)
				continue
			if key == "missing":
				self.set_missing(value)
				continue
			# Anything else is a constraint
			if key == "constraints" and isinstance(value, dict):
				for k, v in value.items():
					self.set_constraint(k, v)


import whyqd as _w

class TestField:

	def test_create(self):
		f = _w.Field("test_name", "string")
		assert f.name == "test_name"
		assert f.type == "string"
		assert f.validates

	def test_constraints(self):
		f = _w.Field("test_name", "string")
		f.set_constraint("required", True)
		assert f.validates
		assert f.required
		category = [
			{
				"name": "dog",
				"description": "A type of mammal"
			},
			{
				"name": "cat",
				"description": "A different type of mammal"
			},
			{
				"name": "mouse",
				"description": "A small type of mammal"
			}
		]
		f.set_constraint("category", category)
		assert f.constraints["category"] == category

	def test_filters(self):
		f = _w.Field("test_name", "date")
		assert f.validates
		filters = {
			"modifiers": ["LATEST", "AFTER"]
		}
		f.set_constraint("filter", filters)
		filters["field"] = True
		assert f.constraints["filter"] == filters

	def test_all(self):
		# Test string field
		settings = {
			"constraints": {
				"required": True,
				"category": [
					{
						"name": "dog",
						"description": "A type of mammal"
					},
					{
						"name": "cat",
						"description": "A different type of mammal"
					},
					{
						"name": "mouse",
						"description": "A small type of mammal"
					}
				]
			}
		}
		f = _w.Field("test_name", "string", **settings)
		assert f.validates
		settings["name"] = "test_name"
		settings["type"] = "string"
		assert f.settings == settings
		# Test date field
		settings = {
			"constraints": {
				"filter": {
					"field": True,
					"modifiers": ["LATEST", "AFTER"]
				}
			}
		}
		f = _w.Field("test_name", "date", **settings)
		assert f.validates
		settings["name"] = "test_name"
		settings["type"] = "date"
		assert f.settings == settings


"""
Actions
-------

Definitions and validation for actions anchoring structures in methods.

Key functions are:

  - Validate an action exists
  - Specify any modifiers
  - Validate action structure
"""
from copy import deepcopy
import whyqd.common as _c

class Action:

	def __init__(self, **action):
		if not action.get("name") and action.get("validate", True):
			e = "Missing type for Action initialisation"
			TypeError(e)
		if action.get("validate", True):
			valid = False
			for valid_action in _c.get_settings("actions")["fields"]:
				if valid_action["name"] == action["name"]:
					action = valid_action
					valid = True
					break
			if not valid:
				e = "Action `{}` is not a valid term.".format(action["name"])
				raise ValueError(e)
		action.pop("validate", None)
		self.action_settings = deepcopy(action)
		self.name = self.action_settings.get("name")
		self.title = self.action_settings.get("title")
		self.description = self.action_settings.get("description")
		self.default_structure = {
			"field": [f["type"] for f in _c.get_settings("schema")["fields"]],
			"modifier": self.modifier_names
		}

	@property
	def modifier_names(self):
		if "modifiers" in self.action_settings:
			return [modifier["name"] for modifier in self.action_settings["modifiers"]]
		return []

	def has_modifier(self, term):
		for modifier in self.action_settings.get("modifiers", []):
			if modifier["name"] == term:
				return True
		return False

	def get_modifier(self, term):
		for modifier in self.action_settings.get("modifiers", []):
			if modifier["name"] == term:
				return deepcopy(modifier)
		return {}

	def has_valid_structure(self, working_columns, structure):
		"""
		Traverses a list defined by `*structure`, ensuring that the terms conform to that action's
		default structural requirements. Nested structures are permitted. Note that the
		responsibility for digging in to the nested structures lies with the calling function.

		The format for defining a `structure` is as follows::

			[action, column_name, [action, column_name]]

		e.g.::

			["CATEGORISE", "+", ["ORDER", "column_1", "column_2"]]

		A calling function would specify::

			Action.has_valid_structure(working_columns, *structure[1:])

		Parameters
		----------
		working_columns: list
			List of valid terms from the working data columns.
		structure: list
			Each structure list must conform to the requirements for that action. Nested actions
			defined by nested lists.

		Returns
		-------
		bool, True if valid
		"""
		if not structure:
			e = "A structure must contain at least one `field`."
			raise ValueError(e)
		term_set = len(self.action_settings.get("structure", [1]))
		for field in _c.chunks(structure, term_set):
			if len(field) != term_set:
				return False
			for i, term in enumerate(self.action_settings["structure"]):
				if term == "field" and isinstance(field[i], list):
					# The calling function needs to handle recursion through the list
					continue
				if term == "value" and not _c.get_field_type(field[i]):
					# Special case for NEW actions where the field can be anything
					return False
				if term == "field" and field[i] not in working_columns:
					return False
				if term == "modifier" and field[i] not in self.default_structure[term]:
					return False
		return True

	@property
	def settings(self):
		"""
		Action settings

		Returns
		-------
		dict: settings
		"""
		return deepcopy(self.action_settings)

	def validate_structure(self, working_columns, *structure):
		"""
		A recursive function testing method structure and that field names are entities of the list
		of `working_columns`. A method should have fields upon which actions are applied, but each
		field may have sub-fields requiring their own actions.

		Parameters
		----------
		working_columns: list
			Permitted column names from `working_data`
		method: list of dicts
			Each dict an element of a `method`, and must start with an `action`
		"""
		# Analyse requirements for this field
		if len(structure) < 2 or structure[0]["type"] != "action":
			e = "A method must start with an `action` and contain at least one `field`."
			raise ValueError(e)
		action = structure[0]
		# Superficial validation check
		for field in structure[1:]:
			# Validate that the source column names are actually in the `working_columns`
			if field["type"] in self.default_field_types and field["name"] not in working_columns:
				e = "Method field `{}` of type `{}` in action `{}` has invalid column `{}`."
				e = e.format(self._name, self._type, action["name"], field["name"])
				raise ValueError(e)
			# Recurse for any nested fields down the tree
			if field["type"] == "nested":
				try:
					self.validate_structure(working_columns, *field["action"])
				except KeyError:
					e = "Method field `{}` of type `{}` in action `{}` has missing nested action."
					e = e.format(self._name, self._type, action["name"])
					raise KeyError(e)
			if not field:
				e = "Method field `{}` of type `{}` in action `{}` has void field."
				e = e.format(self._name, self._type, action["name"])
				raise ValueError(e)
		# Validation of specific action structure
		if action["name"] not in ["ORDER_NEW", "ORDER_OLD", "CALCULATE", "CATEGORISE"]:
			return True
		if action["name"] in ["ORDER_NEW", "ORDER_OLD"]:
			# Requires sets of 3 terms: field, +, date_field
			term_set = 3
		if action["name"] in ["CALCULATE", "CATEGORISE"]:
			# Requires sets of 2 terms: + or -, field
			term_set = 2
		valid = True
		for field in _c.chunks(structure[1:], term_set):
			if len(field) != term_set:
				valid = False
			if valid and action["name"] in ["ORDER_NEW", "ORDER_OLD"]:
				if ((field[0]["type"] not in self.default_field_types + ["nested"]) or
					(field[1]["name"] != "+") or
					(field[2]["type"] not in self.default_field_types + ["nested"])):
					valid = False
			if valid and action["name"] in ["CALCULATE", "CATEGORISE"]:
				if ((field[0]["type"] != "modifier") or
					(field[1]["type"] not in self.default_field_types + ["nested"])):
					valid = False
			if valid and action["name"] == "CATEGORISE":
				if not field[1].get("constraints", {}).get("category"):
					valid = False
				if ((field[0]["name"] == "-") and
					(field[1]["constraints"]["category"] != [True, False])):
					valid = False
				if ((field[0]["name"] == "+") and
					(len(field[1]["constraints"]["category"]) !=
					 len(set(field[1]["constraints"]["category"])))):
					valid = False
			if not valid:
				e = "Method field `{}` of type `{}` in action `{}` has invalid structure."
				e = e.format(self._name, self._type, action["name"], field["name"])
				raise ValueError(e)
		return True

	def validates(self, working_columns):
		"""
		Test action `structure` for structure (every action starts with an action `type`), and that
		columns, categories and filters (depending on what is supplied) are valid.

		Parameters
		----------
		working_columns: list
			Permitted column names from `working_data`

		Raises
		------
		ValueError if working data not provided, if names not permitted, or if action not valid

		Returns
		-------
		Bool, True if valid
		"""
		if "structure" not in self.settings:
			e = "Method field `{}` of type `{}` has no `methods`.".format(self._name, self._type)
			raise ValueError(e)
		self.validate_structure(working_columns, self.settings["structure"])
		# Need to validate the filters
		return super().validates


from pathlib import Path
import os

import whyqd as _w
import whyqd.common as _c

method_name = "/data/test_method.json"
schema_name = "/data/test_schema.json"
DIRECTORY = str(Path(__file__).resolve().parent)
METHOD_SOURCE = DIRECTORY + method_name
SCHEMA_SOURCE = DIRECTORY + schema_name
METHOD = _c.load_json(METHOD_SOURCE)
SCHEMA = _c.load_json(SCHEMA_SOURCE)
INPUT_DATA = [DIRECTORY + "/data/raw_E06000044_014_0.XLSX",
			  DIRECTORY + "/data/raw_E06000044_014_1.XLSX",
			  DIRECTORY + "/data/raw_E06000044_014_2.XLSX"]

class TestMethod:

	def test_create(self, tmp_path):
		DIRECTORY = str(tmp_path) + "/"
		_c.check_path(DIRECTORY)
		method = _w.Method(SCHEMA_SOURCE, directory=DIRECTORY, input_data=INPUT_DATA)
		method.set_details(name="test_method")
		method.save()
		test_source = DIRECTORY + "test_method.json"
		method = _w.Method(test_source, directory=DIRECTORY)
		# Reversed
		assert METHOD["input_data"][2]["checksum"] == method.input_data[0]["checksum"]
		assert METHOD["input_data"][1]["checksum"] == method.input_data[1]["checksum"]
		assert METHOD["input_data"][0]["checksum"] == method.input_data[2]["checksum"]

	def test_merge(self, tmp_path):
		DIRECTORY = str(tmp_path) + "/"
		_c.check_path(DIRECTORY)
		method = _w.Method(SCHEMA_SOURCE, directory=DIRECTORY, input_data=INPUT_DATA)
		method.set_details(name="test_method")
		oak = [
			{
				"id": method.input_data[2]["id"],
				"key": "Property ref no"
			},
			{
				"id": method.input_data[1]["id"],
				"key": "Property Reference Number"
			},
			{
				"id": method.input_data[0]["id"],
				"key": "Property Reference Number"
			}
		]
		method.merge(order_and_key=oak)
		method.save(overwrite=True)
		test_source = DIRECTORY + "test_method.json"
		method = _w.Method(test_source, directory=DIRECTORY)
		assert METHOD["working_data"]["checksum"] == method.working_data["checksum"]
		assert method.validate_merge

	def test_structure(self, tmp_path):
		DIRECTORY = str(tmp_path) + "/"
		_c.check_path(DIRECTORY)
		method = _w.Method(SCHEMA_SOURCE, directory=DIRECTORY, input_data=INPUT_DATA)
		method.set_details(name="test_method")
		oak = [
			{
				"id": method.input_data[2]["id"],
				"key": "Property ref no"
			},
			{
				"id": method.input_data[1]["id"],
				"key": "Property Reference Number"
			},
			{
				"id": method.input_data[0]["id"],
				"key": "Property Reference Number"
			}
		]
		method.merge(order_and_key=oak)
		structure = {
			"la_code": ["NEW", "E06000044"],
			"ba_ref": ["RENAME", "Property ref no"],
			"prop_ba_rates": ["ORDER", "Current Rateable Value_x", "Current Rateable Value_y", "Current Rateable Value"],
			"occupant_name": ["ORDER", "Primary Liable party name_x", "Primary Liable party name_y", "Primary Liable party name"],
			"postcode": ["ORDER", "Full Property Address_x", "Full Property Address_y", "Full Property Address"],
			"occupation_state": ["CATEGORISE",
								 "+", 'Current Property Exemption Code',
								 "+", "Current Relief Type"],
			"occupation_state_date": ["ORDER_NEW",
									  "Current Prop Exemption Start Date", "+", "Current Prop Exemption Start Date",
									  "Current Relief Award Start Date", "+", "Current Relief Award Start Date",
									  "Account Start date_x", "+", "Account Start date_x",
									  "Account Start date_y", "+", "Account Start date_y"],
			"occupation_state_reliefs": ["CATEGORISE",
										 "+", 'Current Property Exemption Code',
										 "+", "Current Relief Type"]
		}
		method.set_structure(**structure)
		method.save(overwrite=True)
		test_source = DIRECTORY + "test_method.json"
		method = _w.Method(test_source, directory=DIRECTORY)
		method.validate_structure

	def test_category(self, tmp_path):
		DIRECTORY = str(tmp_path) + "/"
		_c.check_path(DIRECTORY)
		method = _w.Method(SCHEMA_SOURCE, directory=DIRECTORY, input_data=INPUT_DATA)
		method.set_details(name="test_method")
		oak = [
			{
				"id": method.input_data[0]["id"],
				"key": "Property Reference Number"
			},
			{
				"id": method.input_data[1]["id"],
				"key": "Property Reference Number"
			},
			{
				"id": method.input_data[2]["id"],
				"key": "Property ref no"
			}
		]
		method.merge(order_and_key=oak)
		structure = {
			"la_code": ["NEW", "E06000044"],
			"ba_ref": ["RENAME", "Property ref no"],
			"prop_ba_rates": ["ORDER", "Current Rateable Value_x", "Current Rateable Value_y", "Current Rateable Value"],
			"occupant_name": ["ORDER", "Primary Liable party name_x", "Primary Liable party name_y", "Primary Liable party name"],
			"postcode": ["ORDER", "Full Property Address_x", "Full Property Address_y", "Full Property Address"],
			"occupation_state": ["CATEGORISE",
								 "+", 'Current Property Exemption Code',
								 "+", "Current Relief Type"],
			"occupation_state_date": ["ORDER_NEW",
									  "Current Prop Exemption Start Date", "+", "Current Prop Exemption Start Date",
									  "Current Relief Award Start Date", "+", "Current Relief Award Start Date",
									  "Account Start date_x", "+", "Account Start date_x",
									  "Account Start date_y", "+", "Account Start date_y"],
			"occupation_state_reliefs": ["CATEGORISE",
										 "+", 'Current Property Exemption Code',
										 "+", "Current Relief Type"]
		}
		method.set_structure(**structure)
		category = {
			"occupation_state_reliefs": {
				"small_business": [
					'Small Business Relief England::Current Relief Type',
					'Sbre Extension For 12 Months::Current Relief Type',
					'Supporting Small Business Relief::Current Relief Type'
					],
				"enterprise_zone": ['INDUSTRIAL::Current Property Exemption Code'],
				"vacancy": [
					'EPRN::Current Property Exemption Code',
					'EPRI::Current Property Exemption Code',
					'VOID::Current Property Exemption Code',
					'Empty Property Rate Non-Industrial::Current Relief Type',
					'Empty Property Rate Industrial::Current Relief Type',
					'EPCH::Current Property Exemption Code',
					'LIQUIDATE::Current Property Exemption Code',
					'DECEASED::Current Property Exemption Code',
					'PROHIBITED::Current Property Exemption Code',
					'BANKRUPT::Current Property Exemption Code',
					'Empty Property Rate Charitable::Current Relief Type'
					],
				"retail": ['Retail Discount::Current Relief Type'],
				"exempt": [
					'C::Current Property Exemption Code',
					'LOW RV::Current Property Exemption Code',
					'LAND::Current Property Exemption Code'
					],
				"other": [
					'Sports Club (Registered CASC)::Current Relief Type',
					'Mandatory::Current Relief Type'
					]
				},
			"occupation_state": {
				"false": [
					'EPRN::Current Property Exemption Code',
					'EPRI::Current Property Exemption Code',
					'VOID::Current Property Exemption Code',
					'Empty Property Rate Non-Industrial::Current Relief Type',
					'Empty Property Rate Industrial::Current Relief Type',
					'EPCH::Current Property Exemption Code',
					'LIQUIDATE::Current Property Exemption Code',
					'DECEASED::Current Property Exemption Code',
					'PROHIBITED::Current Property Exemption Code',
					'BANKRUPT::Current Property Exemption Code',
					'Empty Property Rate Charitable::Current Relief Type'
					]
				}
			}
		method.set_category(**category)
		method.save(overwrite=True)
		test_source = DIRECTORY + "test_method.json"
		method = _w.Method(test_source, directory=DIRECTORY)
		test_category = {}
		for field_name in method.all_field_names:
			if method.field(field_name).get("category"):
				test_category[field_name] = method.category(field_name).get("assigned")
		assert category.keys() == test_category.keys()
		for key in category:
			assert category[key].keys() == test_category[key].keys()
			for k in category[key]:
				assert set(category[key][k]) == set(test_category[key][k])
		assert method.validate_category
		method.set_filter("occupation_state_date", "AFTER", "2010-01-01", "ba_ref")
		method.save(overwrite=True)
		test_source = DIRECTORY + "test_method.json"
		method = _w.Method(test_source, directory=DIRECTORY)
		assert method.validate_filter
		method.transform(overwrite_output=True)
		method.save(overwrite=True)
		method = _w.Method(test_source, directory=DIRECTORY)
		assert method.validate_transform


import setuptools

DESCRIPTION = "data wrangling simplicity, complete audit transparency, and at speed "
LONG_DESCRIPTION = """
**whyqd** provides an intuitive method for restructuring messy data to conform to a standardised
metadata schema. It supports data managers and researchers looking to rapidly, and continuously,
normalise any messy spreadsheets using a simple series of steps. Once complete, you can import
wrangled data into more complex analytical systems or full-feature wrangling tools.

It aims to get you to the point where you can perform automated data munging prior to
committing your data into a database, and no further. It is built on Pandas, and plays well with
existing Python-based data-analytical tools. Each raw source file will produce a json schema and
method file which defines the set of actions to be performed to produce refined data, and a
destination file validated against that schema.

**whyqd** ensures complete audit transparency by saving all actions performed to restructure
your input data to a separate json-defined methods file. This permits others to scrutinise your
approach, validate your methodology, or even use your methods to import data in production.

Once complete, a method file can be shared, along with your input data, and anyone can
import **whyqd** and validate your method to verify that your output data is the product of these
inputs.

## Why use it?

If all you want to do is test whether your source data are even useful, spending days or weeks
slogging through data restructuring could kill a project. If you already have a workflow and
established software which includes Python and pandas, having to change your code every time your
source data changes is really, really frustrating.

There are two complex and time-consuming parts to preparing data for analysis: social, and technical.

The social part requires multi-stakeholder engagement with source data-publishers, and with
destination database users, to agree structural metadata. Without any agreement on data publication
formats or destination structure, you are left with the tedious frustration of manually wrangling
each independent dataset into a single schema.

**whyqd** allows you to get to work without requiring you to achieve buy-in from anyone or change
your existing code.

## Wrangling process

  - Create, update or import a data schema which defines the destination data structure;
  - Create a new method and associate it with your schema and input data source/s;
  - Assign a foreign key column and (if required) merge input data sources;
  - Structure input data fields to conform to the requriements for each schema field;
  - Assign categorical data identified during structuring;
  - Transform and filter input data to produce a final destination data file;
  - Share your data and a citation;
"""
DISTNAME = "whyqd"
LICENSE = "BSD"
AUTHOR = "Gavin Chait"
EMAIL = "gchait@whythawk.com"
URL = "http://whyqd.com"
DOWNLOAD_URL = ""
PROJECT_URLS = {
    "Bug Tracker": "https://github.com/whythawk/whyqd/issues/",
    "Documentation": "https://whyqd.readthedocs.io/",
    "Source Code": "https://github.com/whythawk/whyqd/",
}
CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Environment :: Console",
    "Intended Audience :: Science/Research",
	"License :: OSI Approved :: BSD License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Topic :: Scientific/Engineering"
]

setuptools.setup(
    name=DISTNAME,
    version="0.1.1",
    author=AUTHOR,
    author_email=EMAIL,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license=LICENSE,
	url=URL,
	download_url=DOWNLOAD_URL,
	project_urls=PROJECT_URLS,
	classifiers=CLASSIFIERS,
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
)


from pathlib import Path

import whyqd as _w
import whyqd.common as _c

filename = "/data/test_schema.json"
source = str(Path(__file__).resolve().parent) + filename
data = _c.load_json(source)

class TestSchema:

	def test_create(self):
		s = _w.Schema()
		field = {
			"name": "test_field",
			"type": "string",
			"constraints": {
				"required": True,
				"category": [
					{
						"name": "dog",
						"description": "A type of mammal"
					},
					{
						"name": "cat",
						"description": "A different type of mammal"
					},
					{
						"name": "mouse",
						"description": "A small type of mammal"
					}
				]
			}
		}
		s.set_details(name="test_schema")
		s.set_field(**field)
		assert s.validates

	def test_load(self):
		s = _w.Schema(source)
		assert s.validates
		assert s.settings == data

	def test_build(self):
		s = _w.Schema()
		details = {
			"name": data["name"],
			"title": data["title"],
			"description": data["description"]
		}
		s.set_details(**details)
		for field in data["fields"]:
			s.set_field(**field)
		assert s.validates
		assert s.settings == data


import sys
try:
	assert sys.version_info >= (3,7)
except AssertionError:
	from whyqd.schema import Field, Schema
	from whyqd.method import Action, Method
else:
	from .schema import Field, Schema
	from .method import Action, Method